"""
    Copyright 2020 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

MIN_BUCKET_NAME_LEN = 3
MAX_BUCKET_NAME_LEN = 63

ALL_REGIONS = ['us-east-1', 'us-east-2', 'us-west-1', 'us-west-2', 'sa-east-1',
               'ca-central-1', 'eu-west-1', 'eu-central-1', 'eu-west-2',
               'eu-west-3', 'ap-northeast-1', 'ap-northeast-2',
               'ap-southeast-1', 'ap-southeast-2', 'ap-south-1', 'eu-north-1']

REQUIRED = 'required'
VALIDATOR = 'validator'

PROJECT_PATH_CFG = 'project_path'
ACCOUNT_ID_CFG = 'account_id'
REGION_CFG = 'region'
LAMBDAS_ALIASES_NAME_CFG = 'lambdas_alias_name'

AWS_ACCESS_KEY_ID_CFG = 'aws_access_key_id'
AWS_SECRET_ACCESS_KEY_CFG = 'aws_secret_access_key'
DEPLOY_TARGET_BUCKET_CFG = 'deploy_target_bucket'
PROJECTS_MAPPING_CFG = 'build_projects_mapping'
RESOURCES_SUFFIX_CFG = 'resources_suffix'
RESOURCES_PREFIX_CFG = 'resources_prefix'

PYTHON_BUILD_TOOL_NAME = 'python'
NODE_BUILD_TOOL_NAME = 'node'
MVN_BUILD_TOOL_NAME = 'mvn'
ALLOWED_BUILD_TOOLS = [PYTHON_BUILD_TOOL_NAME,
                       MVN_BUILD_TOOL_NAME,
                       NODE_BUILD_TOOL_NAME]

REQUIRED_PARAM_ERROR = 'The required key {} is missing'


class ConfigValidator:

    def __init__(self, config_dict) -> None:
        self._config_dict = config_dict
        self._fields_validators_mapping = {
            PROJECT_PATH_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_project_path},
            ACCOUNT_ID_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_account_id},
            REGION_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_region},
            DEPLOY_TARGET_BUCKET_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_bundle_bucket_name},
            PROJECTS_MAPPING_CFG: {
                REQUIRED: True,
                VALIDATOR: self._validate_project_mapping},
            AWS_ACCESS_KEY_ID_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_access_key},
            AWS_SECRET_ACCESS_KEY_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_aws_secret_access_key},
            RESOURCES_PREFIX_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_resources_prefix_suffix},
            RESOURCES_SUFFIX_CFG: {
                REQUIRED: False,
                VALIDATOR: self._validate_resources_prefix_suffix}
        }

    def validate(self):
        error_messages = {}
        for key, value in self._config_dict.items():
            validation_rules = self._fields_validators_mapping.get(key)
            if not validation_rules:
                raise AssertionError(
                    f'There is no validator for the configuration field {key}')

            is_required = validation_rules.get(REQUIRED)
            if is_required:
                if not value:
                    error_messages[key] = REQUIRED_PARAM_ERROR.format(key)
                    continue

            validator_func = validation_rules.get(VALIDATOR)
            validation_errors = validator_func(key, value)
            if validation_errors:
                error_messages[key] = validation_errors
        return error_messages

    def _validate_project_path(self, key, value):
        str_error = self._assert_value_is_str(key, value)
        if str_error:
            return [str_error]
        errors = []
        if len(value) == 0:
            errors.append(f'{key} must not be empty')
        if not os.path.exists(value):
            errors.append(f'The path {value} specified in {key} must exist')
        return errors

    @staticmethod
    def _validate_account_id(key, value):
        errors = []
        try:
            int(value)
        except TypeError as e:
            errors.append(f'{key} must be int, not {type(value)}')
            return errors
        if len(str(value)) != 12:
            errors.append(f'{key} must be a 12-digit number')
        return errors

    def _validate_region(self, key, value):
        str_error = self._assert_value_is_str(key, value)
        if str_error:
            return [str_error]
        if value not in ALL_REGIONS:
            return [
                f'{key} value must be one of {ALL_REGIONS}, but is {value}'
            ]

    def _validate_bundle_bucket_name(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

        errors = []
        # check min length
        if len(value) < MIN_BUCKET_NAME_LEN or len(
                value) > MAX_BUCKET_NAME_LEN:
            errors.append(f'The length of {key} must be between '
                          f'{MIN_BUCKET_NAME_LEN} and {MAX_BUCKET_NAME_LEN} '
                          f'characters long')
        return errors

    def _validate_project_mapping(self, key, value):
        errors = []
        if type(value) is not dict:
            errors.append(f'{key} must be type of dict')
            return errors
        project_path = self._config_dict.get(PROJECT_PATH_CFG)
        for key in value.keys():
            if key not in ALLOWED_BUILD_TOOLS:
                errors.append(f'{key} is not supported to be built')
                continue
            for build_key, paths in value.items():
                for path in paths:
                    if not os.path.exists(os.path.join(project_path, path)):
                        errors.append(f'The path in {key}:{build_key} project '
                                      f'mapping does not exists: {path}')
        return errors

    def _validate_aws_access_key(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]
        if len(value) < 16 or len(value) > 128:
            return [
                f'The length of {key} must be in a '
                f'range between 16 and 128 characters']

    def _validate_aws_secret_access_key(self, key, value):
        # the only constraint found
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]

    def _validate_resources_prefix_suffix(self, key, value):
        str_error = self._assert_value_is_str(key=key,
                                              value=value)
        if str_error:
            return [str_error]
        if len(value) > 5:
            return [
                f'The length of {key} must be less or equal to 5 character']

    @staticmethod
    def _assert_value_is_str(key, value):
        if type(value) is not str:
            return f'{key} must be type of string'



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import json
from concurrent.futures import ALL_COMPLETED, ThreadPoolExecutor
from functools import cmp_to_key

from syndicate.commons.log_helper import get_logger
from syndicate.core.build.bundle_processor import (create_deploy_output,
                                                   load_deploy_output,
                                                   load_failed_deploy_output,
                                                   load_meta_resources,
                                                   remove_deploy_output,
                                                   remove_failed_deploy_output)
from syndicate.core.build.helper import _json_serial
from syndicate.core.build.meta_processor import resolve_meta
from syndicate.core.constants import (BUILD_META_FILE_NAME,
                                      CLEAN_RESOURCE_TYPE_PRIORITY,
                                      DEPLOY_RESOURCE_TYPE_PRIORITY,
                                      UPDATE_RESOURCE_TYPE_PRIORITY)
from syndicate.core.helper import exit_on_exception, prettify_json

_LOG = get_logger('syndicate.core.build.deployment_processor')


def get_dependencies(name, meta, resources_dict, resources):
    """ Get dependencies from resources that needed to create them too.

    :type name: str
    :type meta: dict
    :type resources_dict: dict
    :param resources:
    :param resources_dict: resources that will be created {name: meta}
    """
    resources_dict[name] = meta
    if meta.get('dependencies'):
        for dependency in meta.get('dependencies'):
            dep_name = dependency['resource_name']
            dep_meta = resources[dep_name]
            resources_dict[dep_name] = dep_meta
            if dep_meta.get('dependencies'):
                get_dependencies(dep_name, dep_meta, resources_dict, resources)


# todo implement resources sorter according to priority
def _process_resources(resources, handlers_mapping, pass_context=False):
    output = {}
    args = []
    resource_type = None
    try:
        for res_name, res_meta in resources:
            current_res_type = res_meta['resource_type']

            if resource_type is None:
                resource_type = current_res_type

            if current_res_type == resource_type:
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                continue
            elif current_res_type != resource_type:
                _LOG.info('Processing {0} resources ...'.format(resource_type))
                func = handlers_mapping[resource_type]
                response = func(args)  # todo exception may be raised here
                if response:
                    output.update(response)
                del args[:]
                args.append(_build_args(name=res_name,
                                        meta=res_meta,
                                        context=output,
                                        pass_context=pass_context))
                resource_type = current_res_type
        if args:
            _LOG.info('Processing {0} resources ...'.format(resource_type))
            func = handlers_mapping[resource_type]
            response = func(args)
            if response:
                output.update(response)
        return True, output
    except Exception as e:
        _LOG.exception('Error occurred while {0} '
                       'resource creating: {1}'.format(resource_type, str(e)))
        # args list always contains one item here
        return False, update_failed_output(args[0]['name'], args[0]['meta'],
                                           resource_type, output)


def _build_args(name, meta, context, pass_context=False):
    """
    Builds parameters to pass to resource_type handler.
    Default parameters dict consists of name and meta keys.
    If pass_context set to True, parameters dict is extended with 'context' key
    :param name: name of the resource
    :param meta: definition of the resource
    :param context: result of previously deployed resources in scope of current syndicate execution
    :param pass_context: flag. Manages if output will be included to parameters
    :return: prepared parameters to be passed to resource_type handler.
    """
    params = {'name': name, 'meta': meta}
    if pass_context:
        params['context'] = context
    return params


def update_failed_output(res_name, res_meta, resource_type, output):
    from syndicate.core import PROCESSOR_FACADE
    describe_func = PROCESSOR_FACADE.describe_handlers()[resource_type]
    failed_resource_output = describe_func(res_name, res_meta)
    if failed_resource_output:
        if isinstance(failed_resource_output, list):
            for item in failed_resource_output:
                output.update(item)
        else:
            output.update(failed_resource_output)
    return output


def deploy_resources(resources):
    from syndicate.core import PROCESSOR_FACADE
    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.create_handlers())


def update_resources(resources):
    from syndicate.core import PROCESSOR_FACADE
    return _process_resources(
        resources=resources,
        handlers_mapping=PROCESSOR_FACADE.update_handlers(),
        pass_context=True)


def clean_resources(output):
    from syndicate.core import PROCESSOR_FACADE
    args = []
    resource_type = None
    # clean all resources
    for arn, config in output:
        res_type = config['resource_meta']['resource_type']
        if resource_type is None:
            resource_type = res_type

        if res_type == resource_type:
            args.append({'arn': arn, 'config': config})
            continue
        elif res_type != resource_type:
            _LOG.info('Removing {0} resources ...'.format(resource_type))
            func = PROCESSOR_FACADE.remove_handlers()[resource_type]
            func(args)
            del args[:]
            args.append({'arn': arn, 'config': config})
            resource_type = res_type
    if args:
        _LOG.info('Removing {0} resources ...'.format(resource_type))
        func = PROCESSOR_FACADE.remove_handlers()[resource_type]
        func(args)


# todo implement saving failed output
def continue_deploy_resources(resources, failed_output):
    from syndicate.core import PROCESSOR_FACADE
    updated_output = {}
    deploy_result = True
    res_type = None
    try:
        args = []
        resource_type = None
        for res_name, res_meta in resources:
            res_type = res_meta['resource_type']

            if resource_type is None:
                resource_type = res_type

            if res_type == resource_type:
                resource_output = __find_output_by_resource_name(
                    failed_output, res_name)
                args.append(
                    {
                        'name': res_name,
                        'meta': res_meta,
                        'current_configurations': resource_output
                    })
                continue
            elif res_type != resource_type:
                func = PROCESSOR_FACADE.resource_configuration_processor() \
                    .get(resource_type)
                if func:
                    response = func(args)
                    if response:
                        updated_output.update(
                            json.loads(
                                json.dumps(response, default=_json_serial)))
                else:
                    # function to update resource is not present
                    # move existing output for resources to new output
                    __move_output_content(args, failed_output, updated_output)
                del args[:]
                resource_output = __find_output_by_resource_name(
                    failed_output, res_name)
                args.append({
                    'name': res_name,
                    'meta': res_meta,
                    'current_configurations': resource_output
                })
                resource_type = res_type
        if args:
            func = PROCESSOR_FACADE.resource_configuration_processor() \
                .get(resource_type)
            if func:
                response = func(args)
                if response:
                    updated_output.update(
                        json.loads(
                            json.dumps(response, default=_json_serial)))
            else:
                # function to update resource is not present
                # move existing output- for resources to new output
                __move_output_content(args, failed_output, updated_output)
    except Exception as e:
        _LOG.exception(
            'Error occurred while {0} resource creating: {1}'.format(
                res_type, str(e)))
        deploy_result = False

    return deploy_result, updated_output


def __move_output_content(args, failed_output, updated_output):
    for arg in args:
        resource_output = __find_output_by_resource_name(
            failed_output, arg['name'])
        if resource_output:
            updated_output.update(resource_output)


def __find_output_by_resource_name(output, resource_name):
    found_items = {}
    for k, v in output.items():
        if v['resource_name'] == resource_name:
            found_items[k] = v
    return found_items


@exit_on_exception
def create_deployment_resources(deploy_name, bundle_name,
                                deploy_only_resources=None,
                                deploy_only_types=None,
                                excluded_resources=None,
                                excluded_types=None,
                                replace_output=False):
    resources = load_meta_resources(bundle_name)
    # validate_deployment_packages(resources)
    _LOG.info('{0} file was loaded successfully'.format(BUILD_META_FILE_NAME))

    # TODO make filter chain
    if deploy_only_resources:
        resources = dict((k, v) for (k, v) in resources.items() if
                         k in deploy_only_resources)

    if excluded_resources:
        resources = dict((k, v) for (k, v) in resources.items() if
                         k not in excluded_resources)
    if deploy_only_types:
        resources = dict((k, v) for (k, v) in resources.items() if
                         v['resource_type'] in deploy_only_types)

    if excluded_types:
        resources = dict((k, v) for (k, v) in resources.items() if
                         v['resource_type'] not in excluded_types)

    resources = resolve_meta(resources)
    _LOG.debug('Names were resolved')
    _LOG.debug(prettify_json(resources))

    _LOG.debug('Going to create: {0}'.format(prettify_json(resources)))

    # sort resources with priority
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(_compare_deploy_resources))

    _LOG.info('Going to deploy AWS resources')
    success, output = deploy_resources(resources_list)
    if success:
        _LOG.info('AWS resources were deployed successfully')

        # apply dynamic changes that uses ARNs
        _LOG.info('Going to apply dynamic changes')
        _apply_dynamic_changes(resources, output)
        _LOG.info('Dynamic changes were applied successfully')

    _LOG.info('Going to create deploy output')
    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output=output,
                         success=success,
                         replace_output=replace_output)
    _LOG.info('Deploy output for {0} was created.'.format(deploy_name))
    return success


@exit_on_exception
def update_deployment_resources(bundle_name, deploy_name, replace_output=False,
                                update_only_types=None,
                                update_only_resources=None):
    from syndicate.core import PROCESSOR_FACADE
    resources = resolve_meta(load_meta_resources(bundle_name))
    _LOG.debug(prettify_json(resources))

    _LOG.warn(
        'Please pay attention that only the '
        'following resources types are supported for update: {}'.format(
            list(PROCESSOR_FACADE.update_handlers().keys())))

    # TODO make filter chain
    resources = dict((k, v) for (k, v) in resources.items() if
                     v['resource_type'] in
                     PROCESSOR_FACADE.update_handlers().keys())

    if update_only_types:
        resources = dict((k, v) for (k, v) in resources.items() if
                         v['resource_type'] in update_only_types)

    if update_only_resources:
        resources = dict((k, v) for (k, v) in resources.items() if
                         k in update_only_resources)

    _LOG.debug('Going to update the following resources: {0}'.format(
        prettify_json(resources)))
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(_compare_update_resources))
    success, output = _process_resources(
        resources=resources_list,
        handlers_mapping=PROCESSOR_FACADE.update_handlers(),
        pass_context=True)
    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output=output,
                         success=success,
                         replace_output=replace_output)
    return success


def _filter_the_dict(dictionary, callback):
    new_dict = dict()
    for (key, value) in dictionary.items():
        if callback(value):
            new_dict[key] = value
    return new_dict


@exit_on_exception
def remove_deployment_resources(deploy_name, bundle_name,
                                clean_only_resources=None,
                                clean_only_types=None,
                                excluded_resources=None, excluded_types=None):
    output = new_output = load_deploy_output(bundle_name, deploy_name)
    _LOG.info('Output file was loaded successfully')
    filters = [
        lambda v: v['resource_name'] in clean_only_resources,
        lambda v: v['resource_name'] not in excluded_resources,
        lambda v: v['resource_meta']['resource_type'] in clean_only_types,
        lambda v: v['resource_meta']['resource_type'] not in excluded_types
    ]

    for function in filters:
        some_result = _filter_the_dict(new_output, function)
        if some_result:
            new_output = some_result

    # sort resources with priority
    resources_list = list(new_output.items())
    resources_list.sort(key=cmp_to_key(_compare_clean_resources))
    _LOG.debug('Resources to delete: {0}'.format(resources_list))

    _LOG.info('Going to clean AWS resources')
    clean_resources(resources_list)
    # remove new_output from bucket
    if output == new_output:
        remove_deploy_output(bundle_name, deploy_name)
    else:
        for key, value in new_output.items():
            output.pop(key)
        create_deploy_output(bundle_name=bundle_name,
                             deploy_name=deploy_name,
                             output=output,
                             success=True,
                             replace_output=True)


@exit_on_exception
def continue_deployment_resources(deploy_name, bundle_name,
                                  deploy_only_resources=None,
                                  deploy_only_types=None,
                                  excluded_resources=None,
                                  excluded_types=None,
                                  replace_output=False):
    output = load_failed_deploy_output(bundle_name, deploy_name)
    _LOG.info('Failed output file was loaded successfully')

    resources = resolve_meta(load_meta_resources(bundle_name))
    _LOG.debug('Names were resolved')
    _LOG.debug(prettify_json(resources))

    # TODO make filter chain
    if deploy_only_resources:
        resources = dict((k, v) for (k, v) in resources.items() if
                         k in deploy_only_resources)

    if excluded_resources:
        resources = dict((k, v) for (k, v) in resources.items() if
                         k not in excluded_resources)
    if deploy_only_types:
        resources = dict((k, v) for (k, v) in resources.items() if
                         v['resource_type'] in deploy_only_types)

    if excluded_types:
        resources = dict((k, v) for (k, v) in resources.items() if
                         v['resource_type'] not in excluded_types)

    # sort resources with priority
    resources_list = list(resources.items())
    resources_list.sort(key=cmp_to_key(_compare_deploy_resources))

    success, updated_output = continue_deploy_resources(resources_list, output)
    _LOG.info('AWS resources were deployed successfully')
    if success:
        # apply dynamic changes that uses ARNs
        _LOG.info('Going to apply dynamic changes')
        _apply_dynamic_changes(resources, updated_output)
        _LOG.info('Dynamic changes were applied successfully')

    # remove failed output from bucket
    remove_failed_deploy_output(bundle_name, deploy_name)
    _LOG.info('Going to create deploy output')
    create_deploy_output(bundle_name=bundle_name,
                         deploy_name=deploy_name,
                         output=updated_output,
                         success=success,
                         replace_output=replace_output)
    return success


@exit_on_exception
def remove_failed_deploy_resources(deploy_name, bundle_name,
                                   clean_only_resources=None,
                                   clean_only_types=None,
                                   excluded_resources=None,
                                   excluded_types=None):
    output = load_failed_deploy_output(bundle_name, deploy_name)
    _LOG.info('Failed output file was loaded successfully')

    # TODO make filter chain
    if clean_only_resources:
        output = dict((k, v) for (k, v) in output.items() if
                      v['resource_name'] in clean_only_resources)

    if excluded_resources:
        output = dict((k, v) for (k, v) in output.items() if
                      v['resource_name'] not in excluded_resources)

    if clean_only_types:
        output = dict((k, v) for (k, v) in output.items() if
                      v['resource_meta']['resource_type'] in clean_only_types)

    if excluded_types:
        output = dict((k, v) for (k, v) in output.items() if
                      v['resource_meta'][
                          'resource_type'] not in excluded_types)

    # sort resources with priority
    resources_list = list(output.items())
    resources_list.sort(key=cmp_to_key(_compare_clean_resources))

    _LOG.info('Going to clean AWS resources')
    clean_resources(resources_list)
    # remove output from bucket
    remove_failed_deploy_output(bundle_name, deploy_name)


def _apply_dynamic_changes(resources, output):
    from syndicate.core import PROCESSOR_FACADE
    pool = ThreadPoolExecutor(max_workers=5)
    futures = []
    for name, meta in resources.items():
        resource_type = meta['resource_type']
        apply_changes = meta.get('apply_changes')
        if apply_changes:
            for apply_item in apply_changes:
                change_type = apply_item['apply_type']
                dependency_name = apply_item['dependency_name']
                res_config = resources.get(dependency_name)
                if not res_config:
                    _LOG.debug('Dependency resource {0} is not found, '
                               'skipping the apply'.format(dependency_name))
                else:
                    dependency_type = res_config['resource_type']
                    func = PROCESSOR_FACADE.resource_identifier() \
                        .get(resource_type)
                    if func:
                        resource_output = __find_output_by_resource_name(
                            output, name)
                        identifier = func(name, resource_output)
                        apply_func = PROCESSOR_FACADE.mapping_applier() \
                            .get(change_type)
                        if apply_func:
                            alias = '#{' + name + '}'
                            f = pool.submit(apply_func, alias, identifier,
                                            apply_item)
                            futures.append(f)
                        else:
                            _LOG.warn('Dynamic apply is not defined '
                                      'for {0} type'.format(change_type))
                    else:
                        _LOG.warn('Resource identifier is not defined '
                                  'for {0} type'.format(dependency_type))
            _LOG.info('Dynamic changes were applied to {0}'.format(name))
    concurrent.futures.wait(futures, timeout=None, return_when=ALL_COMPLETED)


def _compare_deploy_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = DEPLOY_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_clean_resources(first, second):
    first_resource_type = first[-1]['resource_meta']['resource_type']
    second_resource_type = second[-1]['resource_meta']['resource_type']
    first_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = CLEAN_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_update_resources(first, second):
    first_resource_type = first[-1]['resource_type']
    second_resource_type = second[-1]['resource_type']
    first_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[first_resource_type]
    second_res_priority = UPDATE_RESOURCE_TYPE_PRIORITY[second_resource_type]
    return _compare_res(first_res_priority, second_res_priority)


def _compare_res(first_res_priority, second_res_priority):
    if first_res_priority < second_res_priority:
        return -1
    elif first_res_priority > second_res_priority:
        return 1
    else:
        return 0



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import build_description_obj

_LOG = get_logger('syndicate.core.resources.cognito_identity_resource')


class CognitoResource(BaseResource):

    def __init__(self, cognito_conn, account_id, region) -> None:
        self.connection = cognito_conn
        self.account_id = account_id
        self.region = region

    def cognito_resource_identifier(self, name, output=None):
        if output:
            # cognito currently is not located in different regions
            # process only first object
            pool_output = list(output.items())[0][1]
            # find id from the output
            return pool_output['description']['IdentityPoolId']
        return self.connection.if_pool_exists_by_name(name)

    def create_cognito_identity_pool(self, args):
        """ Create Cognito identity pool in pool in sub processes.

        :type args: list
        """
        return self.create_pool(self._create_cognito_identity_pool_from_meta,
                                args)

    def describe_cognito_pool(self, name, meta, pool_id=None):
        if not pool_id:
            pool_id = self.connection.if_pool_exists_by_name(name)
        if not pool_id:
            return
        response = self.connection.describe_identity_pool(pool_id)
        arn = 'arn:aws:cognito-identity:{0}:{1}:identitypool/{2}'.format(
            self.region, self.account_id, pool_id)
        return {
            arn: build_description_obj(response, name, meta)
        }

    @unpack_kwargs
    def _create_cognito_identity_pool_from_meta(self, name, meta):
        """ Create Cognito identity pool for authentication.

        :type name: str
        :type meta: dict
        """
        pool_id = self.connection.if_pool_exists_by_name(name)
        if pool_id:
            _LOG.warn('%s cognito identity pool exists.', name)
            return self.describe_cognito_pool(name=name, meta=meta,
                                              pool_id=pool_id)

        _LOG.info('Creating identity pool %s', name)
        open_id_provider_names = meta.get('open_id_providers', [])
        open_id_arns = ['arn:aws:iam::{0}:oidc-provider/{1}'.format(
            self.account_id, n) for n in open_id_provider_names]
        pool_id = self.connection.create_identity_pool(
            pool_name=name, provider_name=meta.get('provider_name'),
            open_id_connect_provider_arns=open_id_arns)
        auth_role = meta.get('auth_role')
        unauth_role = meta.get('unauth_role')
        if auth_role or unauth_role:
            self.connection.set_role(pool_id, auth_role, unauth_role)
        _LOG.info('Created cognito identity pool %s', pool_id)
        return self.describe_cognito_pool(name=name, meta=meta,
                                          pool_id=pool_id)

    def remove_cognito_identity_pools(self, args):
        self.create_pool(self._remove_cognito_identity_pool, args)

    @unpack_kwargs
    def _remove_cognito_identity_pool(self, arn, config):
        pool_id = config['description']['IdentityPoolId']
        try:
            self.connection.remove_identity_pool(pool_id)
            _LOG.info('Cognito identity pool %s was removed', pool_id)
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                _LOG.warn('Cognito identity pool %s is not found', id)
            else:
                raise e



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.connection import ConnectionProvider
from syndicate.core.resources.api_gateway_resource import ApiGatewayResource
from syndicate.core.resources.cloud_watch_alarm_resource import \
    CloudWatchAlarmResource
from syndicate.core.resources.cloud_watch_resource import CloudWatchResource
from syndicate.core.resources.cognito_resource import CognitoResource
from syndicate.core.resources.dynamo_db_resource import DynamoDBResource
from syndicate.core.resources.ebs_resource import EbsResource
from syndicate.core.resources.ec2_resource import Ec2Resource
from syndicate.core.resources.iam_resource import IamResource
from syndicate.core.resources.kinesis_resource import KinesisResource
from syndicate.core.resources.lambda_resource import LambdaResource
from syndicate.core.resources.s3_resource import S3Resource
from syndicate.core.resources.sns_resource import SnsResource
from syndicate.core.resources.sqs_resource import SqsResource
from syndicate.core.resources.step_functions_resource import \
    StepFunctionResource


class ResourceProvider:
    instance = None

    def __init__(self, config, credentials, sts_conn) -> None:
        self.sts_conn = sts_conn
        if not ResourceProvider.instance:
            ResourceProvider.instance = ResourceProvider.__Resources(
                config=config,
                credentials=credentials
            )

    def sts(self):
        return self.sts_conn

    def __getattr__(self, item):
        return getattr(self.instance, item)

    class __Resources:

        _conn_provider = None

        _cw_alarm_resource = None
        _cw_resource = None
        _sns_resource = None
        _api_gateway_resource = None
        _cognito_resource = None
        _dynamodb_resource = None
        _ebs_resource = None
        _ec2_resource = None
        _iam_resource = None
        _kinesis_resource = None
        _lambda_resource = None
        _s3_resource = None
        _sqs_resource = None
        _step_functions_resource = None

        def __init__(self, config, credentials) -> None:
            self.credentials = credentials
            self._conn_provider = ConnectionProvider(credentials)
            self.config = config

        def cw_alarm(self, region=None):
            if not region:
                region = self.credentials.get('region')
            if not self._cw_alarm_resource:
                self._cw_alarm_resource = CloudWatchAlarmResource(
                    cw_conn=self._conn_provider.cw_metric(region=region),
                    sns_conn=self._conn_provider.sqs()
                )
            return self._cw_alarm_resource

        def cw(self):
            if not self._cw_resource:
                self._cw_resource = CloudWatchResource(
                    cw_events_conn_builder=self._conn_provider.cw_events,
                    account_id=self.config.account_id
                )
            return self._cw_resource

        def sns(self):
            if not self._sns_resource:
                self._sns_resource = SnsResource(
                    conn_provider=self._conn_provider,
                    region=self.credentials.get('region'))
            return self._sns_resource

        def api_gw(self):
            if not self._api_gateway_resource:
                self._api_gateway_resource = ApiGatewayResource(
                    apigw_conn=self._conn_provider.api_gateway(),
                    lambda_res=self.lambda_resource(),
                    account_id=self.config.account_id,
                    region=self.config.region
                )
            return self._api_gateway_resource

        def cognito(self):
            self._cognito_resource = CognitoResource(
                cognito_conn=self._conn_provider.cognito_identity(),
                account_id=self.config.account_id,
                region=self.config.region
            )
            if not self._cognito_resource:
                pass
            return self._cognito_resource

        def dynamodb(self):
            if not self._dynamodb_resource:
                self._dynamodb_resource = DynamoDBResource(
                    dynamodb_conn=self._conn_provider.dynamodb(),
                    cw_alarm_conn=self._conn_provider.cw_metric(),
                    app_as_conn=self._conn_provider.application_autoscaling(),
                    iam_conn=self._conn_provider.iam()
                )
            return self._dynamodb_resource

        def ebs(self):
            if not self._ebs_resource:
                self._ebs_resource = EbsResource(
                    ec2_conn=self._conn_provider.ec2(),
                    iam_conn=self._conn_provider.iam(),
                    ebs_conn=self._conn_provider.beanstalk(),
                    sns_conn=self._conn_provider.sns(),
                    s3_conn=self._conn_provider.s3(),
                    account_id=self.config.account_id,
                    region=self.config.region,
                    deploy_target_bucket=self.config.deploy_target_bucket
                )
            return self._ebs_resource

        def ec2(self):
            if not self._ec2_resource:
                self._ec2_resource = Ec2Resource(
                    ec2_conn=self._conn_provider.ec2(),
                    iam_conn=self._conn_provider.iam(),
                    account_id=self.config.account_id,
                    region=self.config.region
                )
            return self._ec2_resource

        def iam(self):
            if not self._iam_resource:
                self._iam_resource = IamResource(
                    iam_conn=self._conn_provider.iam(),
                    account_id=self.config.account_id,
                    region=self.config.region
                )
            return self._iam_resource

        def kinesis(self):
            if not self._kinesis_resource:
                self._kinesis_resource = KinesisResource(
                    kin_conn=self._conn_provider.kinesis()
                )
            return self._kinesis_resource

        def lambda_resource(self):
            if not self._lambda_resource:
                self._lambda_resource = LambdaResource(
                    lambda_conn=self._conn_provider.lambda_conn(),
                    s3_conn=self._conn_provider.s3(),
                    cw_logs_conn=self._conn_provider.cw_logs(),
                    sns_conn=self.sns(),
                    iam_conn=self._conn_provider.iam(),
                    dynamodb_conn=self._conn_provider.dynamodb(),
                    sqs_conn=self._conn_provider.sqs(),
                    kinesis_conn=self._conn_provider.kinesis(),
                    cw_events_conn=self._conn_provider.cw_events(),
                    region=self.config.region,
                    account_id=self.config.account_id,
                    deploy_target_bucket=self.config.deploy_target_bucket
                )
            return self._lambda_resource

        def s3(self):
            if not self._s3_resource:
                self._s3_resource = S3Resource(
                    s3_conn=self._conn_provider.s3()
                )
            return self._s3_resource

        def sqs(self):
            if not self._sqs_resource:
                self._sqs_resource = SqsResource(
                    sqs_conn_builder=self._conn_provider.sqs,
                    region=self.config.region,
                    account_id=self.config.account_id
                )
            return self._sqs_resource

        def step_functions(self):
            if not self._step_functions_resource:
                self._step_functions_resource = StepFunctionResource(
                    sf_conn=self._conn_provider.step_functions(),
                    iam_conn=self._conn_provider.iam(),
                    cw_events_conn=self._conn_provider.cw_events(),
                    lambda_conn=self._conn_provider.lambda_conn(),
                    region=self.config.region,
                    account_id=self.config.account_id
                )
            return self._step_functions_resource



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
from logging import DEBUG, Formatter, INFO, getLogger

# file output
file_handler = logging.FileHandler(filename='sdct.log')
file_handler.setFormatter(Formatter('%(asctime)s [%(levelname)s]'
                                    '  %(name)s,%(lineno)s  %(message)s'))
file_handler.setLevel(DEBUG)

# console output
console_handler = logging.StreamHandler()
console_handler.setLevel(INFO)
console_handler.setFormatter(Formatter('%(message)s'))

sdct_logger = getLogger("sdct")
# add handlers
sdct_logger.addHandler(console_handler)
sdct_logger.addHandler(file_handler)
sdct_logger.setLevel(DEBUG)

logging.captureWarnings(True)


def get_logger(log_name, level=DEBUG):
    """
    :param level:   CRITICAL = 50
                    ERROR = 40
                    WARNING = 30
                    INFO = 20
                    DEBUG = 10
                    NOTSET = 0
    :type log_name: str
    :type level: int
    """
    module_logger = sdct_logger.getChild(log_name)
    if level:
        module_logger.setLevel(level)
    return module_logger



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

_LOG = get_logger('syndicate.connection.beanstalk_connection')


@apply_methods_decorator(retry)
class BeanstalkConnection(object):
    """ Elastic Beanstalk connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('elasticbeanstalk', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new Elastic beanstalk connection.')

    def create_application(self, app_name, description=''):
        self.client.create_application(
            ApplicationName=app_name,
            Description=description
        )

    def remove_app(self, app_name, terminate_env_force=True):
        return self.client.delete_application(
            ApplicationName=app_name,
            TerminateEnvByForce=terminate_env_force
        )

    def create_environment(self, app_name, env_name, group_name=None,
                           description=None, cname_prefix=None, tier=None,
                           tags=None, version_label=None, template_name=None,
                           solution_stack_name=None, option_settings=None,
                           options_to_remove=None):
        """
        http://docs.aws.amazon.com/elasticbeanstalk/latest/dg/command-options-general.html
        :type app_name: str
        :type env_name: str
        :type group_name: str
        :type description: str
        :type cname_prefix: str
        :type tier: dict
        :param tier: { 'Name': 'string',
                       'Type': 'string',
                       'Version': 'string' }
        :type tags: list of dicts
        :param tags: [{ 'Key': 'string', 'Value': 'string' }]
        :type version_label: str
        :type template_name: str
        :type solution_stack_name: str
        :type option_settings: list of dicts
        :param option_settings: [{
                                    'ResourceName': 'string',
                                    'Namespace': 'string',
                                    'OptionName': 'string',
                                    'Value': 'string'
                                }]
        :type options_to_remove: list of dicts
        :param options_to_remove: [{
                                    'ResourceName': 'string',
                                    'Namespace': 'string',
                                    'OptionName': 'string'
                                    }]
        """
        params = dict(ApplicationName=app_name, EnvironmentName=env_name)
        if group_name:
            params['GroupName'] = group_name
        if description:
            params['Description'] = description
        if cname_prefix:
            params['CNAMEPrefix'] = cname_prefix
        if tier:
            params['Tier'] = tier
        if tags:
            params['Tags'] = tags
        if version_label:
            params['VersionLabel'] = version_label
        if template_name:
            params['TemplateName'] = template_name
        if solution_stack_name:
            params['SolutionStackName'] = solution_stack_name
        if option_settings:
            params['OptionSettings'] = option_settings
        if options_to_remove:
            params['OptionsToRemove'] = options_to_remove
        return self.client.create_environment(**params)

    def get_settings(self, app_name, template_name=None, env_name=None):
        """ You must specify either template_name or a env_name.

        :type app_name: str
        :type template_name: str
        :type env_name: str
        :return: response
        """
        params = dict(ApplicationName=app_name)
        if template_name:
            params['TemplateName'] = template_name
        if env_name:
            params['EnvironmentName'] = env_name
        return self.client.describe_configuration_settings(**params)

    def create_app_version(self, app_name, version_label, s3_bucket, s3_key):
        return self.client.create_application_version(
            ApplicationName=app_name,
            VersionLabel=version_label,
            SourceBundle={
                'S3Bucket': s3_bucket,
                'S3Key': s3_key
            }
        )

    def describe_applications(self, app_names=None):
        """
        :type app_names: list
        :return:
        """
        try:
            if app_names is None:
                app_names = []
            return self.client.describe_applications(
                ApplicationNames=app_names).get('Applications')
        except ClientError as e:
            if 'ResourceNotFoundException' in str(e):
                pass  # valid exception
            else:
                raise e

    def describe_environment_health(self, env_name, env_id=None,
                                    attr_names=None):
        """
        :type env_name: str
        :type env_id: str
        :type attr_names: list
        :param attr_names: 'Status' | 'Color' | 'Causes' | 'ApplicationMetrics'
        | 'InstancesHealth' | 'All' | 'HealthStatus' | 'RefreshedAt'
        :return: response
        """
        params = dict(EnvironmentName=env_name)
        if env_id:
            params['EnvironmentId'] = env_id
        if attr_names:
            params['AttributeNames'] = attr_names
        return self.client.describe_environment_health(**params)

    def deploy_env_version(self, app_name, env_name, version_label):
        return self.client.update_environment(ApplicationName=app_name,
                                              EnvironmentName=env_name,
                                              VersionLabel=version_label)

    def describe_available_solutions_stack_names(self):
        solution_stacks = self.client.list_available_solution_stacks()
        if solution_stacks:
            return solution_stacks['SolutionStacks']



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from json import load

from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import (build_py_package_name,
                                         resolve_bundle_directory)
from syndicate.core.build.validator.mapping import (VALIDATOR_BY_TYPE_MAPPING,
                                                    ALL_TYPES)
from syndicate.core.conf.processor import GLOBAL_AWS_SERVICES
from syndicate.core.constants import (API_GATEWAY_TYPE, ARTIFACTS_FOLDER,
                                      BUILD_META_FILE_NAME, EBS_TYPE,
                                      LAMBDA_CONFIG_FILE_NAME, LAMBDA_TYPE,
                                      RESOURCES_FILE_NAME, RESOURCE_LIST,
                                      IAM_ROLE, LAMBDA_LAYER_TYPE,
                                      S3_PATH_NAME)
from syndicate.core.helper import (build_path, prettify_json,
                                   resolve_aliases_for_string,
                                   write_content_to_file)
from syndicate.core.resources.helper import resolve_dynamic_identifier

DEFAULT_IAM_SUFFIX_LENGTH = 5

_LOG = get_logger('syndicate.core.build.meta_processor')


def validate_deployment_packages(bundle_path, meta_resources):
    package_paths = artifact_paths(meta_resources)
    nonexistent_packages = []
    for package in package_paths:
        package_path = build_path(bundle_path, package)
        if not os.path.exists(package_path):
            nonexistent_packages.append(package_path)

    if nonexistent_packages:
        raise AssertionError('Bundle is not properly configured.'
                             ' Nonexistent deployment packages: '
                             '{0}'.format(prettify_json(nonexistent_packages)))


def artifact_paths(meta_resources):
    return [i for i in
            [_retrieve_package(v) for v in list(meta_resources.values())] if i]


def _retrieve_package(meta):
    s3_path = meta.get(S3_PATH_NAME)
    if s3_path:
        return s3_path


def _check_duplicated_resources(initial_meta_dict, additional_item_name,
                                additional_item):
    """ Match two meta dicts (overall and separated) for duplicates.

    :type initial_meta_dict: dict
    :type additional_item_name: str
    :type additional_item: dict
    """
    if additional_item_name in initial_meta_dict:
        additional_type = additional_item['resource_type']
        initial_item = initial_meta_dict.get(additional_item_name)
        if not initial_item:
            return
        initial_type = initial_item['resource_type']
        if additional_type == initial_type == API_GATEWAY_TYPE:
            # check if APIs have same resources
            for each in list(initial_item['resources'].keys()):
                if each in list(additional_item['resources'].keys()):
                    raise AssertionError(
                        "API '{0}' has duplicated resource '{1}'! Please, "
                        "change name of one resource or remove one.".format(
                            additional_item_name, each))
                    # check is APIs have once described cache configuration
            initial_cache_config = initial_item.get(
                'cluster_cache_configuration')
            additional_cache_config = additional_item.get(
                'cluster_cache_configuration')
            if initial_cache_config and additional_cache_config:
                raise AssertionError(
                    "API '{0}' has duplicated cluster cache configurations. "
                    "Please, remove one cluster cache configuration.".format(
                        additional_item_name)
                )
            if initial_cache_config:
                additional_item[
                    'cluster_cache_configuration'] = initial_cache_config
            # handle responses
            initial_responses = initial_item.get(
                'api_method_responses')
            additional_responses = additional_item.get(
                'api_method_responses')
            if initial_responses and additional_responses:
                raise AssertionError(
                    "API '{0}' has duplicated api method responses "
                    "configurations. Please, remove one "
                    "api method responses configuration.".format(
                        additional_item_name)
                )
            if initial_responses:
                additional_item[
                    'api_method_responses'] = initial_responses
            # handle integration responses
            initial_integration_resp = initial_item.get(
                'api_method_integration_responses')
            additional_integration_resp = additional_item.get(
                'api_method_integration_responses')
            if initial_integration_resp and additional_integration_resp:
                raise AssertionError(
                    "API '{0}' has duplicated api method integration "
                    "responses configurations. Please, remove one "
                    "api method integration responses configuration.".format(
                        additional_item_name)
                )
            if initial_integration_resp:
                additional_item[
                    'api_method_integration_responses'] = initial_integration_resp
            # join items dependencies
            dependencies_dict = {each['resource_name']: each
                                 for each in additional_item['dependencies']}
            for each in initial_item['dependencies']:
                if each['resource_name'] not in dependencies_dict:
                    additional_item['dependencies'].append(each)
            # join items resources
            additional_item['resources'].update(initial_item['resources'])
            # return aggregated API description
            init_deploy_stage = initial_item.get('deploy_stage')
            if init_deploy_stage:
                additional_item['deploy_stage'] = init_deploy_stage
            init_apply = initial_item.get('apply_changes', [])
            add_apply = additional_item.get('apply_changes', [])
            additional_item['apply_changes'] = init_apply + add_apply

            return additional_item

        elif additional_type == initial_type:
            if additional_item == initial_item:
                raise AssertionError(
                    'Warn. Two equals resources descriptions were found! '
                    'Please, remove one of them. Resource name:'
                    ' {0}'.format(additional_item_name))
            else:
                raise AssertionError(
                    "Error! Two resources with equal names were found! Name:"
                    " {0}. Please, rename one of them. First resource: {1}. "
                    "Second resource: {2}".format(additional_item_name,
                                                  initial_item,
                                                  additional_item))


def _populate_s3_path_python_node(meta, bundle_name):
    name = meta.get('name')
    version = meta.get('version')
    if not name or not version:
        raise AssertionError('Lambda config must contain name and version. '
                             'Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name,
                                        build_py_package_name(name, version))


def _populate_s3_path_java(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError('Lambda config must contain deployment_package. '
                             'Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_lambda(meta, bundle_name):
    runtime = meta.get('runtime')
    if not runtime:
        raise AssertionError(
            'Lambda config must contain runtime. '
            'Existing configuration: {0}'.format(prettify_json(meta)))
    resolver_func = RUNTIME_PATH_RESOLVER.get(runtime.lower())
    if resolver_func:
        resolver_func(meta, bundle_name)
    else:
        raise AssertionError(
            'Specified runtime {0} in {1} is not supported. '
            'Supported runtimes: {2}'.format(
                runtime.lower(), meta.get('name'),
                list(RUNTIME_PATH_RESOLVER.keys())))


def _populate_s3_path_lambda_layer(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError(
            'Lambda Layer config must contain deployment_package. '
            'Existing configuration'
            ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path_ebs(meta, bundle_name):
    deployment_package = meta.get('deployment_package')
    if not deployment_package:
        raise AssertionError('Beanstalk_app config must contain '
                             'deployment_package. Existing configuration'
                             ': {0}'.format(prettify_json(meta)))
    else:
        meta[S3_PATH_NAME] = build_path(bundle_name, deployment_package)


def _populate_s3_path(meta, bundle_name):
    resource_type = meta.get('resource_type')
    mapping_func = S3_PATH_MAPPING.get(resource_type)
    if mapping_func:
        mapping_func(meta, bundle_name)


RUNTIME_PATH_RESOLVER = {
    'python2.7': _populate_s3_path_python_node,
    'python3.7': _populate_s3_path_python_node,
    'java8': _populate_s3_path_java,
    'nodejs10.x': _populate_s3_path_python_node,
    'nodejs8.10': _populate_s3_path_python_node
}

S3_PATH_MAPPING = {
    LAMBDA_TYPE: _populate_s3_path_lambda,
    EBS_TYPE: _populate_s3_path_ebs,
    LAMBDA_LAYER_TYPE: _populate_s3_path_lambda_layer
}


def _look_for_configs(nested_files, resources_meta, path, bundle_name):
    """ Look for all config files in project structure. Read content and add
    all meta to overall meta if there is no duplicates. If duplicates found -
    raise AssertionError.

    :type nested_files: list
    :type resources_meta: dict
    :type path: str
    """
    for each in nested_files:
        if each.endswith(LAMBDA_CONFIG_FILE_NAME):
            lambda_config_path = os.path.join(path, each)
            _LOG.debug('Processing file: {0}'.format(lambda_config_path))
            with open(lambda_config_path) as data_file:
                lambda_conf = load(data_file)

            lambda_name = lambda_conf['name']
            _LOG.debug('Found lambda: {0}'.format(lambda_name))
            _populate_s3_path(lambda_conf, bundle_name)
            res = _check_duplicated_resources(resources_meta, lambda_name,
                                              lambda_conf)
            if res:
                lambda_conf = res
            resources_meta[lambda_name] = lambda_conf

        if each == RESOURCES_FILE_NAME:
            additional_config_path = os.path.join(path, RESOURCES_FILE_NAME)
            _LOG.debug('Processing file: {0}'.format(additional_config_path))
            with open(additional_config_path, encoding='utf-8') as json_file:
                deployment_resources = load(json_file)
            for resource_name in deployment_resources:
                _LOG.debug('Found resource ' + resource_name)
                resource = deployment_resources[resource_name]
                # check if resource type exists in deployment framework and
                #  has resource_type field
                try:
                    resource_type = resource['resource_type']
                except KeyError:
                    raise AssertionError(
                        "There is not 'resource_type' in {0}".format(
                            resource_name))
                if resource_type not in RESOURCE_LIST:
                    raise KeyError(
                        "You specified new resource type in configuration"
                        " file {0}, but it doesn't have creation function."
                        " Please, add new creation function or change "
                        "resource name with existing one.".format(
                            resource_type))
                _populate_s3_path(resource, bundle_name)
                res = _check_duplicated_resources(resources_meta,
                                                  resource_name, resource)
                if res:
                    resource = res
                resources_meta[resource_name] = resource


# todo validate all required configs
def create_resource_json(project_path, bundle_name):
    """ Create resource catalog json with all resource metadata in project.
    :param project_path: path to the project
    :type bundle_name: name of the bucket subdir
    """
    resources_meta = {}

    for path, _, nested_items in os.walk(project_path):
        # there is no duplicates in single json, because json is a dict

        _look_for_configs(nested_items, resources_meta, path, bundle_name)

    # check if all dependencies were described
    common_validator = VALIDATOR_BY_TYPE_MAPPING[ALL_TYPES]
    for name, meta in resources_meta.items():
        common_validator(resource_meta=meta, all_meta=resources_meta)

        resource_type = meta['resource_type']
        type_validator = VALIDATOR_BY_TYPE_MAPPING.get(resource_type)
        if type_validator:
            type_validator(name, meta)

    return resources_meta


def _resolve_names_in_meta(resources_dict, old_value, new_value):
    if isinstance(resources_dict, dict):
        for k, v in resources_dict.items():
            # if k == old_value:
            #     resources_dict[new_value] = resources_dict.pop(k)
            if isinstance(v, str) and old_value == v:
                resources_dict[k] = v.replace(old_value, new_value)
            elif isinstance(v, str) and old_value in v and v.startswith('arn'):
                resources_dict[k] = v.replace(old_value, new_value)
            else:
                _resolve_names_in_meta(v, old_value, new_value)
    elif isinstance(resources_dict, list):
        for item in resources_dict:
            if isinstance(item, dict):
                _resolve_names_in_meta(item, old_value, new_value)
            elif isinstance(item, str):
                if item == old_value:
                    index = resources_dict.index(old_value)
                    del resources_dict[index]
                    resources_dict.append(new_value)


def create_meta(project_path, bundle_name):
    # create overall meta.json with all resource meta info
    meta_path = build_path(project_path, ARTIFACTS_FOLDER,
                           bundle_name)
    _LOG.info("Bundle path: {0}".format(meta_path))
    overall_meta = create_resource_json(project_path=project_path,
                                        bundle_name=bundle_name)
    bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
    write_content_to_file(bundle_dir, BUILD_META_FILE_NAME, overall_meta)


def resolve_meta(overall_meta):
    from syndicate.core import CONFIG
    iam_suffix = _resolve_iam_suffix(iam_suffix=CONFIG.iam_suffix)
    if CONFIG.aliases:
        for key, value in CONFIG.aliases.items():
            name = '${' + key + '}'
            overall_meta = resolve_dynamic_identifier(name, str(value),
                                                      overall_meta)
            _LOG.debug('Resolved meta was created')
    _LOG.debug(prettify_json(overall_meta))
    # get dict with resolved prefix and suffix in meta resources
    # key: current_name, value: resolved_name
    resolved_names = {}
    for name, res_meta in overall_meta.items():
        resource_type = res_meta['resource_type']
        if resource_type in GLOBAL_AWS_SERVICES:
            resolved_name = resolve_resource_name(
                resource_name=name,
                prefix=CONFIG.resources_prefix,
                suffix=CONFIG.resources_suffix)
            # add iam_suffix to IAM role only if it is specified in config file
            if resource_type == IAM_ROLE and iam_suffix:
                resolved_name = resolved_name + iam_suffix
            if name != resolved_name:
                resolved_names[name] = resolved_name
    _LOG.debug('Going to resolve names in meta')
    _LOG.debug('Resolved names mapping: {0}'.format(str(resolved_names)))
    for current_name, resolved_name in resolved_names.items():
        overall_meta[resolved_name] = overall_meta.pop(current_name)
        _resolve_names_in_meta(overall_meta, current_name, resolved_name)
    return overall_meta


def resolve_resource_name(resource_name, prefix=None, suffix=None):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, prefix), suffix)


def resolve_resource_name_by_data(resource_name, resource_prefix,
                                  resource_suffix):
    return _resolve_suffix_name(
        _resolve_prefix_name(resource_name, resource_prefix), resource_suffix)


def _resolve_prefix_name(resource_name, resource_prefix):
    if resource_prefix:
        return resolve_aliases_for_string(resource_prefix) + resource_name
    return resource_name


def _resolve_suffix_name(resource_name, resource_suffix):
    if resource_suffix:
        return resource_name + resolve_aliases_for_string(resource_suffix)
    return resource_name


def _resolve_iam_suffix(suffix_len=DEFAULT_IAM_SUFFIX_LENGTH, iam_suffix=None):
    """
    This method adds additional suffix to iam roles.
    The suffix could be passed to the method. Otherwise it will be generated
    as a random string with the combination of lowercase letters.
    """
    if not iam_suffix:
        return None
    if suffix_len > DEFAULT_IAM_SUFFIX_LENGTH:
        raise AssertionError(
            'Additional suffix for IAM roles should be maximum'
            '{0} symbols in length. Provided: {1}'.format(
                DEFAULT_IAM_SUFFIX_LENGTH, suffix_len))

    # check and use provided
    provided_max_len = DEFAULT_IAM_SUFFIX_LENGTH
    if len(iam_suffix) > provided_max_len:
        raise AssertionError(
            'Provided additional suffix for IAM roles should be maximum'
            '{0} symbols in length. Provided len: {1}; Suffix: {2}'.format(
                provided_max_len, len(iam_suffix), iam_suffix))
    return iam_suffix



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.commons.log_helper import get_logger
from syndicate.core.build.validator.dynamodb_validator import validate_dynamodb
from syndicate.core.constants import LAMBDA_CONFIG_FILE_NAME, DYNAMO_TABLE_TYPE

ALL_TYPES = 'all_types'

_LOG = get_logger('validator')


def common_validate(resource_meta, all_meta):
    dependencies = resource_meta.get('dependencies')
    if dependencies:
        for dependency in resource_meta['dependencies']:
            dependency_name = dependency.get('resource_name')
            if dependency_name not in list(all_meta.keys()):
                err_mess = ("One of resource dependencies wasn't "
                            "described: {0}. Please, describe this "
                            "resource in {1} if it is Lambda or in "
                            "deployment_resources.json"
                            .format(dependency_name,
                                    LAMBDA_CONFIG_FILE_NAME))
                raise AssertionError(err_mess)


# validation customization
VALIDATOR_BY_TYPE_MAPPING = {
    ALL_TYPES: common_validate,
    DYNAMO_TABLE_TYPE: validate_dynamodb
}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json

from syndicate.core.conf.validator import LAMBDAS_ALIASES_NAME_CFG
from syndicate.core.generators import (_alias_variable,
                                       FILE_LAMBDA_HANDLER_NODEJS)

POLICY_LAMBDA_BASIC_EXECUTION = "lambda-basic-execution"

LAMBDA_ROLE_NAME_PATTERN = '{0}-role'  # 0 - lambda_name

SRC_MAIN_JAVA = '/src/main/java'
FILE_POM = '/pom.xml'
CANCEL_MESSAGE = 'Creating of {} has been canceled.'

JAVA_LAMBDA_HANDLER_CLASS = """package {java_package_name};

import com.amazonaws.services.lambda.runtime.Context;
import com.amazonaws.services.lambda.runtime.RequestHandler;
import com.syndicate.deployment.annotations.lambda.LambdaHandler;

import java.util.HashMap;
import java.util.Map;

@LambdaHandler(lambdaName = "{lambda_name}",
	roleName = "{lambda_role_name}",
	isPublishVersion = true,
	aliasName = "${lambdas_alias_name}"
)
public class {lambda_class_name} implements RequestHandler<Object, Map<String, Object>> {

	public Map<String, Object> handleRequest(Object request, Context context) {
		System.out.println("Hello from lambda");
		Map<String, Object> resultMap = new HashMap<String, Object>();
		resultMap.put("statusCode", 200);
		resultMap.put("body", "Hello from Lambda");
		return resultMap;
	}
}
"""

JAVA_ROOT_POM_TEMPLATE = """<?xml version="1.0" encoding="UTF-8"?>
<project xmlns="http://maven.apache.org/POM/4.0.0"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd">
    <modelVersion>4.0.0</modelVersion>

    <groupId>{project_name}-group</groupId>
    <artifactId>{project_name}</artifactId>
    <version>1.0.0</version>

    <properties>
        <maven-shade-plugin.version>3.2.0</maven-shade-plugin.version>
        <deployment-configuration-annotations.version>1.5.8</deployment-configuration-annotations.version>
    </properties>

    <dependencies>
        <!-- AWS dependencies-->
        <dependency>
            <groupId>com.amazonaws</groupId>
            <artifactId>aws-lambda-java-core</artifactId>
            <version>1.2.0</version>
        </dependency>
        <!--Syndicate annotations-->
        <dependency>
            <groupId>net.sf.aws-syndicate</groupId>
            <artifactId>deployment-configuration-annotations</artifactId>
            <version>${deployment-configuration-annotations.version}</version>
        </dependency>
    </dependencies>

    <build>
        <plugins>
            <plugin>
                <groupId>net.sf.aws-syndicate</groupId>
                <artifactId>deployment-configuration-maven-plugin</artifactId>
                <version>${deployment-configuration-annotations.version}</version>
                <configuration>
                    <packages>
                        <!--packages to scan-->
                    </packages>
                    <fileName>${project.name}-${project.version}.jar</fileName>
                </configuration>
                <executions>
                    <execution>
                        <id>generate-config</id>
                        <phase>compile</phase>
                        <inherited>false</inherited>
                        <goals>
                            <goal>gen-deployment-config</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-shade-plugin</artifactId>
                <version>${maven-shade-plugin.version}</version>
                <configuration>
                    <filters>
                        <filter>
                            <artifact>*:*</artifact>
                            <excludes>
                                <exclude>META-INF/*.SF</exclude>
                                <exclude>META-INF/*.DSA</exclude>
                                <exclude>META-INF/*.RSA</exclude>
                            </excludes>
                        </filter>
                    </filters>
                    <createDependencyReducedPom>false</createDependencyReducedPom>
                </configuration>
                <executions>
                    <execution>
                        <phase>package</phase>
                        <goals>
                            <goal>shade</goal>
                        </goals>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>

</project>
"""

PYTHON_LAMBDA_HANDLER_TEMPLATE = """import json


def lambda_handler(event, context):
    # TODO implement
    return {
        'statusCode': 200,
        'body': json.dumps('Hello from Lambda!')
    }
"""

NODEJS_LAMBDA_HANDLER_TEMPLATE = """exports.handler = async (event) => {
    // TODO implement
    const response = {
        statusCode: 200,
        body: JSON.stringify('Hello from Lambda!'),
    };
    return response;
};
"""


def _stringify(dict_content):
    return json.dumps(dict_content, indent=2)


def _generate_python_node_lambda_config(lambda_name, lambda_relative_path):
    return _stringify({
        'version': '1.0',
        'name': lambda_name,
        'func_name': 'handler.lambda_handler',
        'resource_type': 'lambda',
        'iam_role_name': LAMBDA_ROLE_NAME_PATTERN.format(lambda_name),
        'runtime': 'python3.7',
        'memory': 128,
        'timeout': 100,
        'lambda_path': lambda_relative_path,
        'dependencies': [],
        'event_sources': [],
        'env_variables': {},
        'publish_version': True,
        'alias': _alias_variable(LAMBDAS_ALIASES_NAME_CFG)
    })


def _generate_nodejs_node_lambda_config(lambda_name, lambda_relative_path):
    return _stringify({
        'version': '1.0',
        'name': lambda_name,
        'func_name': 'index.handler',
        'resource_type': 'lambda',
        'iam_role_name': LAMBDA_ROLE_NAME_PATTERN.format(lambda_name),
        'runtime': 'nodejs10.x',
        'memory': 128,
        'timeout': 100,
        'lambda_path': lambda_relative_path,
        'dependencies': [],
        'event_sources': [],
        'env_variables': {},
        'publish_version': True,
        'alias': _alias_variable(LAMBDAS_ALIASES_NAME_CFG)
    })


def _generate_package_nodejs_lambda(lambda_name):
    return _stringify({
        "name": lambda_name,
        "version": "1.0.0",
        "description": "",
        "main": FILE_LAMBDA_HANDLER_NODEJS[1:],
        "scripts": {},
        "author": "",
        "license": "ISC",
        "dependencies": {
        }
    })


def _generate_package_lock_nodejs_lambda(lambda_name):
    return _stringify({
        "name": lambda_name,
        "version": "1.0.0",
        "lockfileVersion": 1,
        "requires": True,
        "dependencies": {}
    })


def _get_lambda_default_policy():
    return _stringify({
        POLICY_LAMBDA_BASIC_EXECUTION: {
            'policy_content': {
                "Statement": [
                    {
                        "Action": [
                            "logs:CreateLogGroup",
                            "logs:CreateLogStream",
                            "logs:PutLogEvents",
                            "dynamodb:Get*",
                            "dynamodb:Put*",
                            "dynamodb:Describe*",
                            "xray:PutTraceSegments",
                            "xray:PutTelemetryRecords",
                            "kms:Decrypt"
                        ],
                        "Effect": "Allow",
                        "Resource": "*"
                    }
                ],
                "Version": "2012-10-17"},
            "resource_type": "iam_policy"
        }
    })


def _generate_lambda_role_config(role_name, stringify=True):
    role_content = {
        role_name: {
            "predefined_policies": [],
            "principal_service": "lambda",
            "custom_policies": [
                POLICY_LAMBDA_BASIC_EXECUTION
            ],
            "resource_type": "iam_role",
            "allowed_accounts": [
                "${account_id}"
            ]
        }
    }
    return _stringify(role_content) if stringify else role_content



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import zipfile
from contextlib import closing
from datetime import datetime, date

from syndicate.core.constants import ARTIFACTS_FOLDER
from syndicate.core.helper import build_path


def build_py_package_name(lambda_name, lambda_version):
    return '{0}-{1}.zip'.format(lambda_name, lambda_version)


def zip_dir(basedir, name):
    assert os.path.isdir(basedir)
    with closing(zipfile.ZipFile(name, "w", zipfile.ZIP_DEFLATED)) as z:
        for root, dirs, files in os.walk(basedir, followlinks=True):
            for fn in files:
                absfn = os.path.join(root, fn)
                zfn = absfn[len(basedir) + len(os.sep):]
                z.write(absfn, zfn)


def _json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def resolve_all_bundles_directory():
    from syndicate.core import CONF_PATH
    return build_path(CONF_PATH, ARTIFACTS_FOLDER)


def resolve_bundle_directory(bundle_name):
    return build_path(resolve_all_bundles_directory(), bundle_name)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import retry
from syndicate.core.helper import prettify_json, unpack_kwargs
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             resolve_dynamic_identifier)

_LOG = get_logger('syndicate.core.resources.iam_resource')


class IamResource(BaseResource):

    def __init__(self, iam_conn, account_id, region) -> None:
        self.iam_conn = iam_conn
        self.account_id = account_id
        self.region = region

    def remove_policies(self, args):
        self.create_pool(self._remove_policy, args)

    @unpack_kwargs
    def _remove_policy(self, arn, config):
        policy_name = config['resource_name']
        try:
            self.iam_conn.remove_policy(arn)
            _LOG.info('IAM policy %s was removed.', policy_name)
        except ClientError as e:
            error_code = e.response['Error']['Code']
            if error_code == 'NoSuchEntity':
                _LOG.warn('IAM policy %s is not found', policy_name)
            elif error_code == 'DeleteConflict':
                _LOG.warn('Cannot remove %s policy, it is attached.',
                          policy_name)
            else:
                raise e

    def remove_roles(self, args):
        self.create_pool(self._remove_role, args)

    @unpack_kwargs
    def _remove_role(self, arn, config):
        role_name = config['resource_name']
        try:
            attached_policies = self.iam_conn.get_role_attached_policies(
                role_name)
            if attached_policies:
                _LOG.debug('Detaching policies from role %s', role_name)
                for item in attached_policies:
                    self.iam_conn.detach_policy(role_name, item.arn)
            inline_policies = self.iam_conn.get_role_inline_policies(role_name)
            if inline_policies:
                _LOG.debug('Deleting inline policies from role %s', role_name)
                for item in inline_policies:
                    item.delete()
            instance_profiles = self.iam_conn.get_instance_profiles_for_role(
                role_name)
            if instance_profiles:
                for each in instance_profiles:
                    self.iam_conn.remove_role_from_instance_profile(role_name,
                                                                    each[
                                                                        'InstanceProfileName'])
            self.iam_conn.remove_role(role_name)
            _LOG.info('IAM role %s was removed.', role_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntity':
                _LOG.warn('IAM role %s is not found ', role_name)
            else:
                raise e

    @retry
    def _remove_instance_profile(self, profile_name):
        try:
            self.iam_conn.remove_instance_profile(profile_name)
        except ClientError as e:
            if e.response['Error']['Code'] == 'NoSuchEntity':
                _LOG.info('Instance profile %s is not found ', profile_name)
            else:
                _LOG.error('Cant remove instance profile: %s.', profile_name)
                raise e
        _LOG.info('Instance profile %s removed.', profile_name)

    def create_policies(self, args):
        return self.create_pool(job=self._create_policy_from_meta,
                                parameters=args)

    def create_roles(self, args):
        return self.create_pool(job=self._create_role_from_meta,
                                parameters=args)

    @unpack_kwargs
    def _create_policy_from_meta(self, name, meta):
        arn = self._build_policy_arn(name)
        response = self.iam_conn.get_policy(arn)
        if response:
            _LOG.warn('IAM policy %s exists.', name)
            return self.describe_policy(name=name, meta=meta)
        policy_content = meta['policy_content']
        self.iam_conn.create_custom_policy(name, policy_content)
        _LOG.info('Created IAM policy %s.', name)
        return self.describe_policy(name=name, meta=meta)

    def _build_policy_arn(self, name):
        return 'arn:aws:iam::{0}:policy/{1}'.format(self.account_id, name)

    def describe_policy(self, name, meta, response=None):
        arn = self._build_policy_arn(name)
        if not response:
            response = self.iam_conn.get_policy(arn)
        if response:
            del response['Arn']
            return {
                arn: build_description_obj(response, name, meta)
            }

    @unpack_kwargs
    def _create_role_from_meta(self, name, meta):
        response = self.iam_conn.get_role(name)
        if response:
            _LOG.warn('IAM role %s exists.', name)
            return self.describe_role(name=name, meta=meta, response=response)
        custom_policies = meta.get('custom_policies', [])
        predefined_policies = meta.get('predefined_policies', [])
        policies = set(custom_policies + predefined_policies)
        allowed_accounts = meta.get('allowed_accounts', [])
        principal_service = meta.get('principal_service')
        instance_profile = meta.get('instance_profile')
        external_id = meta.get('external_id')
        trust_rltn = meta.get('trusted_relationships')
        if principal_service and '{region}' in principal_service:
            principal_service = principal_service.format(region=self.region)
        response = self.iam_conn.create_custom_role(
            role_name=name,
            allowed_account=list(allowed_accounts),
            allowed_service=principal_service,
            external_id=external_id,
            trusted_relationships=trust_rltn)
        if instance_profile:
            try:
                self.iam_conn.create_instance_profile(name)
            except ClientError as e:
                if 'EntityAlreadyExists' in str(e):
                    _LOG.warn('Instance profile %s exists', name)
                else:
                    raise e
            self.iam_conn.add_role_to_instance_profile(name, name)
            # attach policies
        if policies:
            for policy in policies:
                arn = self.iam_conn.get_policy_arn(policy)
                if not arn:
                    raise AssertionError(f'Can not get policy arn: {policy}')
                self.iam_conn.attach_policy(name, arn)
        else:
            raise AssertionError('There are no policies for role: %s.', name)
        _LOG.info('Created IAM role %s.', name)
        return self.describe_role(name=name, meta=meta, response=response)

    def describe_role(self, name, meta, response=None):
        if not response:
            response = self.iam_conn.get_role(role_name=name)
        arn = response['Arn']
        del response['Arn']
        return {
            arn: build_description_obj(response, name, meta)
        }

    def apply_trusted_to_role(self, name, value, apply_config):
        trusted = apply_config['trusted_relationships']
        role_name = apply_config['dependency_name']
        resolved_trusted = resolve_dynamic_identifier(name, value, trusted)
        self.iam_conn.update_assume_role_policy_document(
            role_name=role_name,
            document=prettify_json(resolved_trusted))

    def apply_policy_content(self, name, value, apply_config):
        policy_content = apply_config['policy_content']
        policy_name = apply_config['dependency_name']
        resolved_policy_content = resolve_dynamic_identifier(name, value,
                                                             policy_content)
        policy_arn = 'arn:aws:iam::{0}:policy/{1}'.format(self.account_id,
                                                          policy_name)
        self.iam_conn.create_policy_version(
            policy_arn=policy_arn,
            policy_document=prettify_json(resolved_policy_content),
            set_as_default=True)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

import yaml
from configobj import ConfigObj
from validate import Validator, VdtTypeError

from syndicate.commons.log_helper import get_logger
from syndicate.core.conf.validator import (PROJECT_PATH_CFG, REGION_CFG,
                                           DEPLOY_TARGET_BUCKET_CFG,
                                           ACCOUNT_ID_CFG,
                                           PROJECTS_MAPPING_CFG,
                                           AWS_ACCESS_KEY_ID_CFG,
                                           RESOURCES_PREFIX_CFG,
                                           RESOURCES_SUFFIX_CFG,
                                           AWS_SECRET_ACCESS_KEY_CFG,
                                           ALL_REGIONS, ALLOWED_BUILD_TOOLS,
                                           ConfigValidator)
from syndicate.core.constants import (DEFAULT_SEP, IAM_POLICY, IAM_ROLE,
                                      S3_BUCKET_TYPE)

CONFIG_FILE_NAME = 'syndicate.yml'
ALIASES_FILE_NAME = 'syndicate_aliases.yml'

LEGACY_CONFIG_FILE_NAME = 'sdct.conf'
LEGACY_ALIASES_FILE_NAME = 'sdct_aliases.conf'
DEFAULT_SECTION_NAME = 'default'

_LOG = get_logger('core.conf.config_holder')

GLOBAL_AWS_SERVICES = {IAM_ROLE, IAM_POLICY, S3_BUCKET_TYPE}

REQUIRED_PARAMETERS = {
    PROJECT_PATH_CFG: 'string(min=1)',
    REGION_CFG: "region_func",
    DEPLOY_TARGET_BUCKET_CFG: 'string(min=3, max=63)',
    ACCOUNT_ID_CFG: 'account_func',
    PROJECTS_MAPPING_CFG: 'project_func'
}

ALL_CONFIG_PARAMETERS = [
    AWS_ACCESS_KEY_ID_CFG, AWS_SECRET_ACCESS_KEY_CFG,
    RESOURCES_PREFIX_CFG, RESOURCES_SUFFIX_CFG
]
ALL_CONFIG_PARAMETERS.extend(REQUIRED_PARAMETERS.keys())

ERROR_MESSAGE_MAPPING = {
    PROJECT_PATH_CFG: 'cannot be empty',
    REGION_CFG: "is invalid. Valid options: " + str(ALL_REGIONS),
    DEPLOY_TARGET_BUCKET_CFG: 'length must be between 3 and 63 characters',
    ACCOUNT_ID_CFG: 'must be 12-digit number',
    PROJECTS_MAPPING_CFG: "must be as a mapping of build tool to project "
                          "path, separated by ';'. Build tool name "
                          "and project path should be separated by ':'."
                          " Allowed build "
                          "tools values: " + str(ALLOWED_BUILD_TOOLS),
    RESOURCES_PREFIX_CFG: 'length must be less than or equal to 5',
    RESOURCES_SUFFIX_CFG: 'length must be less than or equal to 5'
}


def _region(value):
    value = value.lower()
    if not isinstance(value, str):
        raise VdtTypeError(value)
    if value not in ALL_REGIONS:
        raise VdtTypeError(value)
    return value


def _account(value):
    if len(value) != 12:
        raise VdtTypeError(value)
    try:
        int(value)
    except:
        raise VdtTypeError(value)
    return value


def _project_mapping(value):
    if value is '' or None:
        return ''  # valid case if you have no projects to build
    mappings = value.split(';')
    for mapping in mappings:
        items = mapping.split(':')
        if len(items) != 2:
            raise VdtTypeError(value)
        if items[0] not in ALLOWED_BUILD_TOOLS:
            raise VdtTypeError(value)
    return value


class ConfigHolder:
    def __init__(self, dir_path):
        con_path = os.path.join(dir_path, CONFIG_FILE_NAME)
        if os.path.isfile(con_path):
            self._init_yaml_config(dir_path=dir_path, con_path=con_path)
        else:
            self._init_ini_config(dir_path=dir_path)

    def _init_yaml_config(self, dir_path, con_path):
        config_content = load_yaml_file_content(file_path=con_path)
        if config_content:
            validator = ConfigValidator(config_content)
            errors = validator.validate()
            if errors:
                raise AssertionError(f'The following error occurred '
                                     f'while {con_path} parsing: {errors}')
        self._config_dict = config_content

        aliases_path = os.path.join(dir_path, ALIASES_FILE_NAME)
        aliases_content = load_yaml_file_content(file_path=aliases_path)
        self._aliases = aliases_content

    def _init_ini_config(self, dir_path):
        con_path = os.path.join(dir_path, LEGACY_CONFIG_FILE_NAME)
        if not os.path.isfile(con_path):
            raise AssertionError(
                'sdct.conf does not exist inside %s folder' % dir_path)
        self._config_path = con_path
        self._config_dict = ConfigObj(con_path,
                                      configspec=REQUIRED_PARAMETERS)
        self._validate_ini()
        alias_path = os.path.join(dir_path, LEGACY_ALIASES_FILE_NAME)
        if not os.path.exists(alias_path):
            _LOG.warn('sdct_aliases.conf does not exist '
                      'inside %s folder' % dir_path)
        else:
            self._aliases = ConfigObj(alias_path)

    def _validate_ini(self):
        # building a validator
        validator = Validator({
            'region_func': _region,
            'account_func': _account,
            'project_func': _project_mapping
        })
        # validate
        param_valid_dict = self._config_dict.validate(validator=validator)
        if not param_valid_dict:
            raise Exception(f'Error while parsing {self._config_path}')
        # check non-required parameters
        prefix_value = self._config_dict.get(RESOURCES_PREFIX_CFG)
        if prefix_value:
            if len(prefix_value) > 5:
                if not isinstance(param_valid_dict, dict):
                    param_valid_dict = {RESOURCES_PREFIX_CFG: False}
                else:
                    param_valid_dict[RESOURCES_PREFIX_CFG] = False

        suffix_value = self._config_dict.get(RESOURCES_SUFFIX_CFG)
        if suffix_value:
            if len(suffix_value) > 5:
                if not isinstance(param_valid_dict, dict):
                    param_valid_dict = {RESOURCES_SUFFIX_CFG: False}
                else:
                    param_valid_dict[RESOURCES_SUFFIX_CFG] = False

        # processing results
        if isinstance(param_valid_dict, dict):
            messages = ''
            for key, value in param_valid_dict.items():
                if not value:
                    messages += '\n{0} {1}'.format(key,
                                                   ERROR_MESSAGE_MAPPING[key])

            if messages:
                raise Exception('Configuration is invalid. ' + messages)

    def _resolve_variable(self, variable_name):
        return self._config_dict.get(variable_name)

    @property
    def project_path(self):
        return path_resolver(self._resolve_variable(PROJECT_PATH_CFG))

    @property
    def account_id(self):
        return str(self._resolve_variable(ACCOUNT_ID_CFG))

    @property
    def access_role(self):
        return self._resolve_variable('access_role')

    @property
    def session_duration(self):
        duration = self._resolve_variable('session_duration')
        if duration:
            return int(self._resolve_variable('session_duration'))

    @property
    def aws_access_key_id(self):
        return self._resolve_variable(AWS_ACCESS_KEY_ID_CFG)

    @property
    def aws_secret_access_key(self):
        return self._resolve_variable(AWS_SECRET_ACCESS_KEY_CFG)

    @property
    def region(self):
        return self._resolve_variable(REGION_CFG)

    @property
    def deploy_target_bucket(self):
        return self._resolve_variable(DEPLOY_TARGET_BUCKET_CFG)

    # mapping build tool : paths to project
    @property
    def build_projects_mapping(self):
        mapping_value = self._resolve_variable(PROJECTS_MAPPING_CFG)
        if type(mapping_value) == dict:
            return mapping_value
        if mapping_value:
            mapping_dict = {}
            for i in mapping_value.split(';'):
                key = i.split(':')[0]
                value = i.split(':')[1]
                list_values = mapping_dict.get(key)
                if list_values:
                    list_values.append(path_resolver(value))
                else:
                    mapping_dict[key] = [path_resolver(value)]
            return mapping_dict

    @property
    def resources_prefix(self):
        prefix = self._resolve_variable(RESOURCES_PREFIX_CFG)
        if prefix is None:
            return ''
        else:
            return prefix

    @property
    def resources_suffix(self):
        suffix = self._resolve_variable(RESOURCES_SUFFIX_CFG)
        if suffix is None:
            return ''
        else:
            return suffix

    @property
    def iam_suffix(self):
        """
        Optional property. It will be included as a ending
        of names for iam_roles.
        :return:
        """
        return self._resolve_variable('iam_suffix')

    @property
    def aliases(self):
        return self._aliases

    def resolve_alias(self, name):
        if self._aliases.get(name):
            return self._aliases[name]


def path_resolver(path):
    return path.replace('\\', DEFAULT_SEP).replace('//', DEFAULT_SEP)


def load_yaml_file_content(file_path):
    if not os.path.isfile(file_path):
        raise AssertionError(f'There is no file by path: {file_path}')
    with open(file_path, 'r') as yaml_file:
        return yaml.load(yaml_file, Loader=yaml.FullLoader)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os

from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import resolve_bundle_directory
from syndicate.core.build.runtime.java import assemble_java_mvn_lambdas
from syndicate.core.build.runtime.nodejs import assemble_node_lambdas
from syndicate.core.build.runtime.python import assemble_python_lambdas

RUNTIME_JAVA_8 = 'java8'
RUNTIME_NODEJS = 'nodejs10.x'
RUNTIME_PYTHON = 'pythonX'

SUPPORTED_RUNTIMES = [
    RUNTIME_JAVA_8,
    RUNTIME_NODEJS,
    RUNTIME_PYTHON
]

RUNTIME_TO_BUILDER_MAPPING = {
    RUNTIME_JAVA_8: assemble_java_mvn_lambdas,
    RUNTIME_NODEJS: assemble_node_lambdas,
    RUNTIME_PYTHON: assemble_python_lambdas
}

_LOG = get_logger('syndicate.core.build.artifact_processor')


def assemble_artifacts(bundle_name, project_path, runtime):
    if runtime not in SUPPORTED_RUNTIMES:
        raise AssertionError(
            'Runtime {} is not supported. '
            'Currently available runtimes:{}'.format(runtime,
                                                     SUPPORTED_RUNTIMES))

    bundle_dir = resolve_bundle_directory(bundle_name=bundle_name)
    if not os.path.exists(bundle_dir):
        os.makedirs(bundle_dir)
    _LOG.debug('Target directory: {0}'.format(bundle_dir))

    assemble_func = RUNTIME_TO_BUILDER_MAPPING.get(runtime)
    if not assemble_func:
        raise AssertionError(
            'There is no assembler for the runtime {}'.format(runtime))
    assemble_func(project_path=project_path,
                  bundles_dir=bundle_dir)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import concurrent
import glob
import json
import os
import shutil
import threading
from concurrent.futures import ALL_COMPLETED
from concurrent.futures.thread import ThreadPoolExecutor
from distutils import dir_util

from syndicate.commons.log_helper import get_logger
from syndicate.core.build.helper import build_py_package_name, zip_dir
from syndicate.core.conf.processor import path_resolver
from syndicate.core.constants import (LAMBDA_CONFIG_FILE_NAME,
                                      REQ_FILE_NAME, LOCAL_REQ_FILE_NAME,
                                      DEFAULT_SEP)
from syndicate.core.helper import (build_path, unpack_kwargs, execute_command,
                                   prettify_json)
from syndicate.core.resources.helper import validate_params

_LOG = get_logger('python_runtime_assembler')

_PY_EXT = "*.py"


def assemble_python_lambdas(project_path, bundles_dir):
    from syndicate.core import CONFIG
    project_base_folder = os.path.basename(os.path.normpath(project_path))
    project_abs_path = build_path(CONFIG.project_path, project_path)
    _LOG.info('Going to process python project by path: {0}'.format(
        project_abs_path))
    executor = ThreadPoolExecutor(max_workers=5)
    futures = []
    for root, sub_dirs, files in os.walk(project_abs_path):
        for item in files:
            if item.endswith(LAMBDA_CONFIG_FILE_NAME):
                _LOG.info('Going to build artifact in: {0}'.format(root))
                arg = {
                    'item': item,
                    'project_base_folder': project_base_folder,
                    'project_path': project_path,
                    'root': root,
                    'target_folder': bundles_dir
                }
                futures.append(executor.submit(_build_python_artifact, arg))
    concurrent.futures.wait(futures, return_when=ALL_COMPLETED)
    executor.shutdown()
    _LOG.info('Python project was processed successfully')


@unpack_kwargs
def _build_python_artifact(item, project_base_folder, project_path, root,
                           target_folder):
    from syndicate.core import CONFIG
    _LOG.debug('Building artifact in {0}'.format(target_folder))
    lambda_config_dict = json.load(open(build_path(root, item)))
    req_params = ['lambda_path', 'name', 'version']
    validate_params(root, lambda_config_dict, req_params)
    lambda_path = path_resolver(lambda_config_dict['lambda_path'])
    lambda_name = lambda_config_dict['name']
    lambda_version = lambda_config_dict['version']
    artifact_name = lambda_name + '-' + lambda_version
    # create folder to store artifacts
    artifact_path = build_path(target_folder, artifact_name)
    _LOG.debug('Artifacts path: {0}'.format(artifact_path))
    os.makedirs(artifact_path)
    _LOG.debug('Folders are created')
    # install requirements.txt content
    # getting file content
    req_path = build_path(root, REQ_FILE_NAME)
    if os.path.exists(req_path):
        _LOG.debug('Going to install 3-rd party dependencies')
        with open(req_path) as f:
            req_list = f.readlines()
        req_list = [path_resolver(r.strip()) for r in req_list]
        _LOG.debug(str(req_list))
        # install dependencies
        for lib in req_list:
            command = 'pip install {0} -t {1}'.format(lib, artifact_path)
            execute_command(command=command)
        _LOG.debug('3-rd party dependencies were installed successfully')

    # install local requirements
    local_req_path = build_path(root, LOCAL_REQ_FILE_NAME)
    if os.path.exists(local_req_path):
        _LOG.debug('Going to install local dependencies')
        _install_local_req(artifact_path, local_req_path, project_base_folder,
                           project_path)
        _LOG.debug('Local dependencies were installed successfully')
    src_path = build_path(CONFIG.project_path, project_path, lambda_path)
    _copy_py_files(src_path, artifact_path)
    package_name = build_py_package_name(lambda_name, lambda_version)
    zip_dir(artifact_path, build_path(target_folder, package_name))
    # remove unused folder
    lock = threading.RLock()
    lock.acquire()
    try:
        shutil.rmtree(artifact_path)
    finally:
        lock.release()
    _LOG.info('Package {0} was created successfully'.format(package_name))


def _install_local_req(artifact_path, local_req_path, project_base_folder,
                       project_path):
    from syndicate.core import CONFIG
    with open(local_req_path) as f:
        local_req_list = f.readlines()
    local_req_list = [path_resolver(r.strip()) for r in local_req_list]
    _LOG.info('Local dependencies: {0}'.format(prettify_json(local_req_list)))
    # copy folders
    for lrp in local_req_list:
        _LOG.info('Processing dependency: {0}'.format(lrp))
        folder_path = build_path(artifact_path, project_base_folder, lrp)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)

        dir_util.copy_tree(build_path(CONFIG.project_path, project_path, lrp),
                           folder_path)
        _LOG.debug('Dependency was copied successfully')

        folders = [r for r in lrp.split(DEFAULT_SEP) if r]
        # process folder from root python project
        folders.insert(0, '')
        i = 0
        temp_path = ''
        while i < len(folders):
            temp_path += DEFAULT_SEP + folders[i]
            src_path = build_path(CONFIG.project_path, project_path,
                                  temp_path)
            dst_path = build_path(artifact_path, project_base_folder,
                                  temp_path)
            _copy_py_files(src_path, dst_path)
            i += 1
        _LOG.debug('Python files from packages were copied successfully')


def _copy_py_files(search_path, destination_path):
    files = glob.iglob(build_path(search_path, _PY_EXT))
    for py_file in files:
        if os.path.isfile(py_file):
            shutil.copy2(py_file, destination_path)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import sys

import click

from syndicate.core import CONF_PATH, initialize_connection
from syndicate.core.build.artifact_processor import (RUNTIME_NODEJS,
                                                     assemble_artifacts,
                                                     RUNTIME_JAVA_8,
                                                     RUNTIME_PYTHON)
from syndicate.core.build.bundle_processor import (create_bundles_bucket,
                                                   load_bundle,
                                                   upload_bundle_to_s3,
                                                   if_bundle_exist)
from syndicate.core.build.deployment_processor import (
    continue_deployment_resources, create_deployment_resources,
    remove_deployment_resources, remove_failed_deploy_resources,
    update_deployment_resources)
from syndicate.core.build.meta_processor import create_meta
from syndicate.core.conf.generator import generate_configuration_files
from syndicate.core.conf.validator import (MVN_BUILD_TOOL_NAME,
                                           PYTHON_BUILD_TOOL_NAME,
                                           NODE_BUILD_TOOL_NAME)
from syndicate.core.decorators import check_deploy_name_for_duplicates
from syndicate.core.groups.generate import generate, GENERATE_GROUP_NAME
from syndicate.core.helper import (check_required_param,
                                   create_bundle_callback,
                                   handle_futures_progress_bar,
                                   resolve_path_callback, timeit,
                                   verify_bundle_callback,
                                   verify_meta_bundle_callback)

INIT_COMMAND_NAME = 'init'
commands_without_config = (
    INIT_COMMAND_NAME,
    GENERATE_GROUP_NAME
)


def _not_require_config(all_params):
    return any(item in commands_without_config for item in all_params)


@click.group(name='syndicate')
@click.version_option()
def syndicate():
    if CONF_PATH:
        click.echo('Path to sdct.conf: ' + CONF_PATH)
        initialize_connection()
    elif _not_require_config(sys.argv):
        pass
    else:
        click.echo('Environment variable SDCT_CONF is not set! '
                   'Please verify that you configured have provided path to '
                   'correct config files or execute `syndicate init` command.')
        sys.exit(1)


@syndicate.command(name=INIT_COMMAND_NAME)
@click.option('--config_path', type=str,
              help='Path to store generated configuration file')
@click.option('--project_path', type=str,
              help='Path to project folder. Default value: working dir')
# account settings
@click.option('--account_id', callback=check_required_param, type=int,
              help='[required] Id of the AWS account where to deploy '
                   'application')
@click.option('--region', type=str, default='us-west-1',
              help='The region that is used to deploy the application')
@click.password_option(
    '--access_key', type=str,
    help='AWS access key id that is used to deploy the application.')  # todo get from .aws/credentials/[default] if not specified
@click.password_option(
    '--secret_key', type=str,
    help='AWS secret key that is used to deploy the application.')  # todo get from .aws/credentials/[default] if not specified
@click.option('--bundle_bucket_name', type=str,
              help='Name of the bucket that is used for uploading artifacts. '
                   'It will be created if specified.')
# build_projects_mapping
@click.option('--python_build_mapping', '-pbm', multiple=True,
              help='List of the folders in a project where '
                   'Python code must be assembled')
@click.option('--java_build_mapping', '-jbm', multiple=True,
              help='List of the folders in a project where '
                   'Java code must be assembled')
@click.option('--nodejs_build_mapping', '-nbm', multiple=True,
              help='List of the folders in a project where '
                   'NodeJS code must be assembled')
# deploy setting
@click.option('--prefix', type=str,
              help='Prefix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}')
@click.option('--suffix', type=str,
              help='Suffix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}')
def init(config_path, project_path, account_id, region, access_key,
         secret_key, bundle_bucket_name, python_build_mapping,
         java_build_mapping, nodejs_build_mapping, prefix, suffix):
    """
    Configures aws-syndicate:
    - generates sdct.conf using the provided parameters;
    - creates default sdct_aliases.conf;
    :param config_path: path where the generated configurations files
        will be stored
    :param project_path: path to a project which is supposed to be deployed
        by aws-syndicate
    :param account_id: id of the AWS account where an application
        will be deployed
    :param region: AWS region name where an application will be deployed
    :param access_key: AWS access key id to access an AWS account specified
        in account_id parameter. If access_key is not specified
        the credentials from /$user/.aws/credentials will be used if any.
        The provided credentials should provide enough permissions to deploy
        the app.
    :param secret_key: AWS secret access key to access an AWS account
        specified in account_id parameter. If secret_key is not specified
        the credentials from /$user/.aws/credentials will be used if any.
        The provided credentials should provide enough permissions to deploy
        the app.
    :param bundle_bucket_name: name of the bucket in specified AWS account
        where application bundles will be stored
    :param python_build_mapping: path[s] to the python parts of an application.
        Example: --python_build_mapping /src/$app_python_module/
    :param java_build_mapping: path[s] to the java parts of an application.
        Example: --java_build_mapping /src/$app_java_module/
    :param nodejs_build_mapping: path[s] to the nodejs parts of an application.
        Example: --nodejs_build_mapping /src/$node_python_module/
    :param prefix: will be added to resource names of types
    iam_role, iam_policy, s3_bucket.
    Resource name pattern: {prefix}resource_name{suffix}
    :param suffix: will be added to resource names of types
    iam_role, iam_policy, s3_bucket.
    Resource name pattern: {prefix}resource_name{suffix}
    :return:
    """
    generate_configuration_files(config_path=config_path,
                                 project_path=project_path,
                                 region=region,
                                 account_id=account_id,
                                 access_key=access_key,
                                 secret_key=secret_key,
                                 bundle_bucket_name=bundle_bucket_name,
                                 python_build_mapping=python_build_mapping,
                                 java_build_mapping=java_build_mapping,
                                 nodejs_build_mapping=nodejs_build_mapping,
                                 prefix=prefix,
                                 suffix=suffix)


@syndicate.command(name='build_bundle')
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--force_upload', is_flag=True, default=False)
@click.pass_context
@timeit
def build_bundle(ctx, bundle_name, force_upload):
    """
    Builds bundle of an application
    :param ctx:
    :param bundle_name: name of the bundle
    :param force_upload: used to override existing bundle
    :return:
    """
    if if_bundle_exist(bundle_name=bundle_name) and not force_upload:
        click.echo('Bundle name \'{0}\' already exists '
                   'in deploy bucket. Please use another bundle '
                   'name or delete the bundle'.format(bundle_name))
        return
    ctx.invoke(build_artifacts, bundle_name=bundle_name)
    ctx.invoke(package_meta, bundle_name=bundle_name)
    ctx.invoke(upload_bundle, bundle_name=bundle_name, force=force_upload)


@syndicate.command(name='deploy')
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--deploy_only_types', multiple=True)
@click.option('--deploy_only_resources', multiple=True)
@click.option('--deploy_only_resources_path', nargs=1)
@click.option('--excluded_resources', multiple=True)
@click.option('--excluded_resources_path', nargs=1)
@click.option('--excluded_types', multiple=True)
@click.option('--continue_deploy', is_flag=True)
@click.option('--replace_output', nargs=1, is_flag=True, default=False)
@check_deploy_name_for_duplicates
@timeit
def deploy(deploy_name, bundle_name, deploy_only_types, deploy_only_resources,
           deploy_only_resources_path, excluded_resources,
           excluded_resources_path, excluded_types, continue_deploy,
           replace_output):
    """
    Deploys infrastructure from the specified bundle
    :param deploy_name: name of the deploy
    :param bundle_name: name of the bundle
    :param deploy_only_types: list of types of the resources to deploy
    :param deploy_only_resources: list of resources names to deploy
    :param deploy_only_resources_path: path to a json file that contains
        a list of resources names to deploy
    :param excluded_resources: names of the resources which must be
        skipped while deploy
    :param excluded_resources_path: path to a json file that contains a list
        of resources names which must be skipped while deploy
    :param excluded_types: list of types of resources which must be
        skipped while deploy
    :param continue_deploy: continues deploy using the failed output.
        Used only after previous deploy fail.
    :param replace_output: flag to override the output file.
        Used if previous output file must be overridden.
    :return:
    """
    click.echo('Command deploy backend')
    click.echo('Deploy name: %s' % deploy_name)
    if deploy_only_resources_path and os.path.exists(
            deploy_only_resources_path):
        deploy_resources_list = json.load(open(deploy_only_resources_path))
        deploy_only_resources = tuple(
            set(deploy_only_resources + tuple(deploy_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
    if continue_deploy:
        deploy_success = continue_deployment_resources(deploy_name,
                                                       bundle_name,
                                                       deploy_only_resources,
                                                       deploy_only_types,
                                                       excluded_resources,
                                                       excluded_types,
                                                       replace_output)

    else:
        deploy_success = create_deployment_resources(deploy_name, bundle_name,
                                                     deploy_only_resources,
                                                     deploy_only_types,
                                                     excluded_resources,
                                                     excluded_types,
                                                     replace_output)
    click.echo('Backend resources were deployed{0}.'.format(
        '' if deploy_success else ' with errors. See deploy output file'))


@syndicate.command(name='update')
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--update_only_types', multiple=True)
@click.option('--update_only_resources', multiple=True)
@click.option('--update_only_resources_path', nargs=1)
@click.option('--replace_output', nargs=1, is_flag=True, default=False)
@check_deploy_name_for_duplicates
@timeit
def update(bundle_name, deploy_name, replace_output,
           update_only_resources,
           update_only_resources_path,
           update_only_types=[]):
    """
    Updates infrastructure from the provided bundle
    :param bundle_name: name of the bundle
    :param deploy_name: name of the deploy
    :param update_only_resources: list of resources names to updated
    :param update_only_resources_path: path to a json file with list of
        resources names to update
    :param update_only_types: optional. List of a resources types to update.
    :param replace_output: flag. If True, existing output file will be replaced
    :return:
    """
    click.echo('Bundle name: {}'.format(bundle_name))
    if update_only_types:
        click.echo('Types to update: {}'.format(list(update_only_types)))
    if update_only_resources:
        click.echo('Resources to update: {}'.format(list(update_only_types)))
    if update_only_resources_path:
        click.echo('Path to list of resources to update: {}'.format(
            update_only_resources_path))

    if update_only_resources_path and os.path.exists(
            update_only_resources_path):
        update_resources_list = json.load(open(update_only_resources_path))
        update_only_resources = tuple(
            set(update_only_resources + tuple(update_resources_list)))
    success = update_deployment_resources(
        bundle_name=bundle_name,
        deploy_name=deploy_name,
        update_only_types=update_only_types,
        update_only_resources=update_only_resources,
        replace_output=replace_output)
    if success:
        return 'Update of resources has been successfully completed'
    return 'Something went wrong during resources update'


@syndicate.command(name='clean')
@timeit
@click.option('--deploy_name', nargs=1, callback=check_required_param)
@click.option('--bundle_name', nargs=1, callback=check_required_param)
@click.option('--clean_only_types', multiple=True)
@click.option('--clean_only_resources', multiple=True)
@click.option('--clean_only_resources_path', nargs=1, type=str)
@click.option('--excluded_resources', multiple=True)
@click.option('--excluded_resources_path', nargs=1, type=str)
@click.option('--excluded_types', multiple=True)
@click.option('--rollback', is_flag=True)
def clean(deploy_name, bundle_name, clean_only_types, clean_only_resources,
          clean_only_resources_path, excluded_resources,
          excluded_resources_path, excluded_types, rollback):
    """
    Cleans the application infrastructure.
    :param deploy_name: name of the deploy
    :param bundle_name: name of the bundle
    :param clean_only_types: list of types that must be cleaned
    :param clean_only_resources: list of names of resources that
        must be cleaned
    :param clean_only_resources_path: path to a json list which contains a list
        of resources names which must be cleaned
    :param excluded_resources: list of the resources names than must be skipped
        while cleaning the application infrastructure
    :param excluded_resources_path: path to a json list which contains a list
        of resources names which must be skipped while cleaning
        the application infrastructure
    :param excluded_types: list of types that must be skipped while cleaning
        the application infrastructure
    :param rollback: used to clean resources that were created in scope of the
        failed deploy
    :return:
    """
    click.echo('Command clean')
    click.echo('Deploy name: %s' % deploy_name)
    if clean_only_types:
        click.echo('Clean only types: %s' % str(clean_only_types))
    if clean_only_resources:
        click.echo('Clean only resources : %s' % clean_only_resources)
    if clean_only_resources_path:
        click.echo(
            'Clean only resources path: %s' % clean_only_resources_path)
    if excluded_resources:
        click.echo('Excluded resources: %s' % str(excluded_resources))
    if excluded_resources_path:
        click.echo('Excluded resources path: %s' % excluded_resources_path)
    if excluded_types:
        click.echo('Excluded types: %s' % str(excluded_types))
    if clean_only_resources_path and os.path.exists(
            clean_only_resources_path):
        clean_resources_list = json.load(open(clean_only_resources_path))
        clean_only_resources = tuple(
            set(clean_only_resources + tuple(clean_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
    if rollback:
        remove_failed_deploy_resources(deploy_name=deploy_name,
                                       bundle_name=bundle_name,
                                       clean_only_resources=clean_only_resources,
                                       clean_only_types=clean_only_types,
                                       excluded_resources=excluded_resources,
                                       excluded_types=excluded_types)
    else:
        remove_deployment_resources(deploy_name=deploy_name,
                                    bundle_name=bundle_name,
                                    clean_only_resources=clean_only_resources,
                                    clean_only_types=clean_only_types,
                                    excluded_resources=excluded_resources,
                                    excluded_types=excluded_types)
    click.echo('AWS resources were removed.')


# =============================================================================


@syndicate.command(name='assemble_java_mvn')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_java_mvn(bundle_name, project_path):
    """
    Builds Java lambdas
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :return:
    """
    click.echo('Command compile java project path: %s' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_JAVA_8)
    click.echo('Java artifacts were prepared successfully.')


@syndicate.command(name='assemble_python')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_python(bundle_name, project_path):
    """
    Builds Python lambdas
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :return:
    """
    click.echo('Command assemble python: project_path: %s ' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_PYTHON)
    click.echo('Python artifacts were prepared successfully.')


@syndicate.command(name='assemble_node')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--project_path', '-path', nargs=1,
              callback=resolve_path_callback)
def assemble_node(bundle_name, project_path):
    """
    Builds NodeJS lambdas
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :return:
    """
    click.echo('Command assemble node: project_path: %s ' % project_path)
    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_NODEJS)
    click.echo('NodeJS artifacts were prepared successfully.')


COMMAND_TO_BUILD_MAPPING = {
    MVN_BUILD_TOOL_NAME: assemble_java_mvn,
    PYTHON_BUILD_TOOL_NAME: assemble_python,
    NODE_BUILD_TOOL_NAME: assemble_node
}


@syndicate.command(name='build_artifacts')
@timeit
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.pass_context
def build_artifacts(ctx, bundle_name):
    """
    Builds the application artifacts
    :param ctx:
    :param bundle_name: name of the bundle to which the artifacts
        will be associated
    :return:
    """
    click.echo('Building artifacts ...')
    from syndicate.core import CONFIG
    if CONFIG.build_projects_mapping:
        for key, values in CONFIG.build_projects_mapping.items():
            for value in values:
                func = COMMAND_TO_BUILD_MAPPING.get(key)
                if func:
                    ctx.invoke(func, bundle_name=bundle_name,
                               project_path=value)
                else:
                    click.echo('Build tool is not supported: %s' % key)
    else:
        click.echo('Projects to be built are not found')


@syndicate.command(name='package_meta')
@timeit
@click.option('--bundle_name', nargs=1, callback=verify_bundle_callback)
def package_meta(bundle_name):
    """
    Generates metadata about the application infrastructure
    :param bundle_name: name of the bundle to generate metadata
    :return:
    """
    from syndicate.core import CONFIG
    click.echo('Package meta, bundle: %s' % bundle_name)
    create_meta(project_path=CONFIG.project_path,
                bundle_name=bundle_name)
    click.echo('Meta was configured successfully.')


@syndicate.command(name='create_deploy_target_bucket')
@timeit
def create_deploy_target_bucket():
    """
    Creates a bucket in AWS account where all bundles will be uploaded
    :return:
    """
    from syndicate.core import CONFIG
    click.echo('Create deploy target sdk: %s' % CONFIG.deploy_target_bucket)
    create_bundles_bucket()
    click.echo('Deploy target bucket was created successfully')


@syndicate.command(name='upload_bundle')
@timeit
@click.option('--bundle_name', nargs=1, callback=verify_meta_bundle_callback)
@click.option('--force', is_flag=True)
def upload_bundle(bundle_name, force=False):
    """
    Uploads bundle from local storage to AWS S3
    :param bundle_name: name of the bundle to upload
    :param force: used if the bundle with the same name as provided
        already exists in an account
    :return:
    """
    click.echo('Upload bundle: %s' % bundle_name)
    if force:
        click.echo('Force upload')
    futures = upload_bundle_to_s3(bundle_name=bundle_name, force=force)
    handle_futures_progress_bar(futures)
    click.echo('Bundle was uploaded successfully')


@syndicate.command(name='copy_bundle')
@click.option('--bundle_name', nargs=1, callback=create_bundle_callback)
@click.option('--src_account_id', '-acc_id', nargs=1,
              callback=check_required_param)
@click.option('--src_bucket_region', '-r', nargs=1,
              callback=check_required_param)
@click.option('--src_bucket_name', '-bucket_name', nargs=1,
              callback=check_required_param)
@click.option('--role_name', '-role', nargs=1,
              callback=check_required_param)
@click.option('--force_upload', is_flag=True, default=False)
@timeit
@click.pass_context
def copy_bundle(ctx, bundle_name, src_account_id, src_bucket_region,
                src_bucket_name, role_name, force_upload):
    """
    Copies the bundle from the specified account-region-bucket to
        account-region-bucket specified in sdct.conf
    :param ctx:
    :param bundle_name: name of the bundle to copy
    :param src_account_id: id of the account where target bundle is stored
    :param src_bucket_region: region of the bucket where target bundle
        is stored
    :param src_bucket_name: name of the bucket where target bundle is stored
    :param role_name: name of the role that is assumed while copying
    :param force_upload: used if the bundle with the same name as provided
        already exists in a target account
    :return:
    """
    click.echo('Copy bundle: %s' % bundle_name)
    click.echo('Bundle name: %s' % bundle_name)
    click.echo('Source account id: %s' % src_account_id)
    click.echo('Source bucket region: %s' % src_bucket_region)
    click.echo('Source bucket name: %s' % src_bucket_name)
    futures = load_bundle(bundle_name, src_account_id, src_bucket_region,
                          src_bucket_name, role_name)
    handle_futures_progress_bar(futures)
    click.echo('Bundle was downloaded successfully')
    ctx.invoke(upload_bundle, bundle_name=bundle_name, force=force_upload)
    click.echo('Bundle was copied successfully')


syndicate.add_command(generate)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)

_LOG = get_logger('syndicate.core.resources.dynamo_db_resource')


class DynamoDBResource(BaseResource):

    def __init__(self, dynamodb_conn, cw_alarm_conn,
                 app_as_conn, iam_conn) -> None:
        self.dynamodb_conn = dynamodb_conn
        self.cw_alarm_conn = cw_alarm_conn
        self.app_autoscaling_conn = app_as_conn
        self.iam_conn = iam_conn

    def create_tables_by_10(self, args):
        """ Only 10 tables can be created, updated or deleted simultaneously.
    
        :type args: list
        """
        response = dict()
        waiters = {}
        start = 0
        end = 8
        while start < len(args):
            tables_to_create = args[start: end]
            for arg_set in tables_to_create:
                name = arg_set['name']
                meta = arg_set['meta']
                response.update(
                    self._create_dynamodb_table_from_meta(name, meta))
                table = self.dynamodb_conn.get_table_by_name(name)
                waiters[table.name] = table.meta.client.get_waiter(
                    'table_exists')
            for table_name in waiters:
                waiters[table_name].wait(TableName=table_name)
            start = end
            end += 9
        return response

    def describe_table(self, name, meta, response=None):
        if not response:
            response = self.dynamodb_conn.describe_table(table_name=name)
        arn = response['TableArn']
        del response['TableArn']
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_stream(self, name, meta):
        response = self.dynamodb_conn.describe_table(meta['table'])
        res_obj = {
            'StreamSpecification': response['StreamSpecification'],
            'LatestStreamLabel': response['LatestStreamLabel']
        }
        arn = response['LatestStreamArn']
        return {
            arn: build_description_obj(res_obj, name, meta)
        }

    def _create_dynamodb_table_from_meta(self, name, meta):
        """ Create Dynamo DB table from meta description after parameter
        validation.
    
        :type name: str
        :type meta: dict
        """
        required_parameters = ['hash_key_name', 'hash_key_type',
                               'read_capacity',
                               'write_capacity']
        validate_params(name, meta, required_parameters)

        res = self.dynamodb_conn.describe_table(name)
        autoscaling_config = meta.get('autoscaling')
        if res:
            _LOG.warn('%s table exists.', name)
            if autoscaling_config:
                res['Autoscaling'] = self._describe_autoscaling(
                    autoscaling_config,
                    name)
            return self.describe_table(name, meta, res)

        self.dynamodb_conn.create_table(
            name, meta['hash_key_name'], meta['hash_key_type'],
            meta.get('sort_key_name'), meta.get('sort_key_type'),
            meta['read_capacity'], meta['write_capacity'],
            global_indexes=meta.get('global_indexes'),
            local_indexes=meta.get('local_indexes'),
            wait=False)
        response = self.dynamodb_conn.describe_table(name)
        if not response:
            raise AssertionError('Table with name {0} has not been created!'
                                 .format(name))
        # enabling stream if present
        stream_view_type = meta.get('stream_view_type')
        if stream_view_type:
            stream = self.dynamodb_conn.get_table_stream_arn(name)
            if stream:
                _LOG.warn('Stream %s exists.', name)
            else:
                try:
                    self.dynamodb_conn.enable_table_stream(name,
                                                           stream_view_type)
                except ClientError as e:
                    # handle specific case for fantom stream enabling
                    if 'ResourceInUseException' in str(e):
                        _LOG.warn('Stream enabling currently in progress,'
                                  ' table: %s', name)
                    else:
                        raise e
        if autoscaling_config:
            _LOG.debug('Found autoscaling configuration for resource %s', name)
            sc_res = self._enable_autoscaling(autoscaling_config, name)
            response['Autoscaling'] = sc_res
        _LOG.info('Created table %s.', name)
        return self.describe_table(name, meta, response)

    def _describe_autoscaling(self, autoscaling_config, name):
        targets = []
        policies = []
        for item in autoscaling_config:
            dimension = item['dimension']
            resource_name = item['resource_name']
            resource_id = self._build_res_id(dimension, resource_name, name)
            sc_targets = self.app_autoscaling_conn.describe_scalable_targets(
                service_namespace='dynamodb',
                resources_ids=[resource_id],
                scalable_dimension=dimension)
            targets.extend(sc_targets)
            autoscaling_policy = item.get('config')
            if autoscaling_policy:
                policy_name = autoscaling_policy['policy_name']
                sc_policies = self.app_autoscaling_conn.describe_scaling_policies(
                    service_namespace='dynamodb', policy_names=[policy_name],
                    resource_id=resource_id, scalable_dimension=dimension)
                policies.extend(sc_policies)
        return {
            'targets': targets,
            'policies': policies
        }

    def _enable_autoscaling(self, autoscaling_config, name):
        targets = []
        policies = []
        for item in autoscaling_config:
            autoscaling_required_parameters = ['resource_name', 'dimension',
                                               'min_capacity', 'max_capacity',
                                               'role_name']
            validate_params(name, item, autoscaling_required_parameters)
            role_name = item['role_name']
            role_arn = self.iam_conn.check_if_role_exists(role_name)
            if role_arn:
                dimension = item['dimension']
                resource_id, sc_targets = self.register_autoscaling_target(
                    dimension,
                    item,
                    role_arn,
                    name)
                targets.extend(sc_targets)
                _LOG.debug('Autoscaling %s is set up for %s', dimension,
                           resource_id)
                autoscaling_policy = item.get('config')
                if autoscaling_policy:
                    policy_name = autoscaling_policy['policy_name']
                    _LOG.debug('Going to set up autoscaling with '
                               'policy %s', policy_name)
                    sc_policies = self.put_autoscaling_policy(
                        autoscaling_policy,
                        dimension,
                        policy_name,
                        resource_id)
                    policies.append(sc_policies)
                    _LOG.debug('Policy %s is set up', policy_name)
            else:
                _LOG.warn('Role %s is not found, skip autoscaling config',
                          role_name)
        return {
            'targets': targets,
            'policies': policies
        }

    def put_autoscaling_policy(self, autoscaling_policy, dimension,
                               policy_name,
                               resource_id):
        target_utilization = autoscaling_policy['target_utilization']
        scale_in_cooldown = autoscaling_policy.get('scale_in_cooldown')
        scale_out_cooldown = autoscaling_policy.get('scale_out_cooldown')
        metric_type = 'DynamoDBWriteCapacityUtilization' \
            if 'Write' in dimension \
            else 'DynamoDBReadCapacityUtilization'
        response = self.app_autoscaling_conn.put_target_scaling_policy(
            policy_name=policy_name, service_namespace='dynamodb',
            resource_id=resource_id, scalable_dimension=dimension,
            target_value=target_utilization,
            predefined_metric_type=metric_type,
            scale_in_cooldown=scale_in_cooldown,
            scale_out_cooldown=scale_out_cooldown)
        return response

    def register_autoscaling_target(self, dimension, item, role_arn,
                                    table_name):
        resource_name = item['resource_name']
        resource_id = self._build_res_id(dimension, resource_name, table_name)
        self.app_autoscaling_conn.register_target(service_namespace='dynamodb',
                                                  resource_id=resource_id,
                                                  scalable_dimension=dimension,
                                                  min_capacity=str(
                                                      item['min_capacity']),
                                                  max_capacity=str(
                                                      item['max_capacity']),
                                                  role_arn=role_arn)
        targets = self.app_autoscaling_conn.describe_scalable_targets(
            service_namespace='dynamodb',
            resources_ids=[resource_id],
            scalable_dimension=dimension)
        return resource_id, targets

    def _build_res_id(self, dimension, resource_name, table_name):
        resource_id = 'table/{0}'.format(table_name) \
            if 'table' in dimension \
            else 'table/{0}/index/{1}'.format(table_name, resource_name)
        return resource_id

    def remove_dynamodb_tables(self, args):
        db_names = [x['config']['resource_name'] for x in args]
        self.dynamodb_conn.remove_tables_by_names(db_names)
        _LOG.info('Dynamo DB tables %s were removed', str(db_names))
        alarm_args = []
        for arg in args:
            autoscaling = arg['config']['description'].get('Autoscaling')
            if autoscaling:
                policies = autoscaling['policies']
                for policy in policies:
                    if policy:
                        alarms = policy.get('Alarms', [])
                        alarm_args.extend(map(lambda x: {
                            'arn': x['AlarmARN'],
                            'config': {'resource_name': x['AlarmName']}
                        }, alarms))

        self.cw_alarm_conn.remove_alarms(alarm_args)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
import pathlib
import click

from syndicate.core.generators.project import (generate_project_structure,
                                               PROJECT_PROCESSORS)
from syndicate.core.generators.lambda_function import (generate_lambda_function)
from syndicate.core.helper import (check_required_param, timeit, OrderedGroup)

GENERATE_GROUP_NAME = 'generate'


@click.group(name=GENERATE_GROUP_NAME, cls=OrderedGroup, chain=True)
def generate():
    """Manages auto-generating"""


@generate.command(name='project')
@click.option('--project_name', nargs=1, callback=check_required_param,
              help='* The project name')
@click.option('--lang', nargs=1, callback=check_required_param,
              help='* The name of programming language that will '
                   'be used in the project',
              type=click.Choice(PROJECT_PROCESSORS))
@click.option('--project_path', nargs=1,
              help='The path where project structure will be created')
@click.pass_context
@timeit
def project(ctx, project_name, lang, project_path):
    """
    Generates project with all the necessary components and in a right
    folders/files hierarchy to start developing in a min.
    :param ctx:
    :param project_name: the project name
    :param lang: name of programming language that will be used in the project
    :param project_path: the path where project structure will be created
    :return:
    """
    click.echo('Project name: {}'.format(project_name))
    click.echo('Language: {}'.format(lang))

    proj_path = pathlib.Path().absolute() if not project_path else project_path
    if not os.access(proj_path, os.X_OK | os.W_OK):
        return ('Incorrect permissions for the provided path {}'.format(
            proj_path))
    click.echo('Project path: {}'.format(proj_path))
    generate_project_structure(project_name=project_name,
                               project_path=proj_path,
                               project_language=lang)


@generate.command(name='lambda')
@click.option('--project_name', nargs=1, callback=check_required_param,
              help='* The project name')
@click.option('--lang', nargs=1, callback=check_required_param,
              help='* The name of programming language that will '
                   'be used in the project',
              type=click.Choice(PROJECT_PROCESSORS))
@click.option('--project_path', nargs=1,
              help='The path where project structure will be created. '
                   '(If not specified - will be used current directory)')
@click.option('--lambda_name', nargs=1, multiple=True, type=str,
              callback=check_required_param,
              help='(multiple) * The lambda function name')
@click.pass_context
@timeit
def lambda_function(ctx, project_name, lang, project_path, lambda_name):
    """
    Generates required environment for lambda function
    :param ctx:
    :param project_name: the project name
    :param lang: name of programming language that will be used in the project
    :param project_path: the path where project structure will be created
    :param lambda_name: the lambda function name (multiple)
    :return:
    """
    proj_path = pathlib.Path().absolute() if not project_path else project_path
    if not os.access(proj_path, os.X_OK | os.W_OK):
        return ('Incorrect permissions for the provided path {}'.format(
            proj_path))

    click.echo('Project name: {}'.format(project_name))
    click.echo('Language: {}'.format(lang))
    click.echo('Project path: {}'.format(proj_path))
    click.echo('Lambda names: {}'.format(str(lambda_name)))
    generate_lambda_function(project_name=project_name,
                             project_path=proj_path,
                             project_language=lang,
                             lambda_names=lambda_name)
